from typing import Any, Dict, Tuple, cast
from unittest.mock import Mock

import numpy as np
import pytest
from numpy.typing import NDArray

from ...models.hybrid_model import HybridModel
from ..processor import QuantumProcessor
from .xgboost_quantum_hybrid import HybridConfig, XGBoostQuantumHybrid


@pytest.fixture
def sample_data():
    """Generate sample data for testing"""
    rng = np.random.default_rng(seed=42)
    features = rng.standard_normal((100, 10))  # 100 samples, 10 features
    labels = rng.integers(0, 2, 100)  # Binary classification
    return features, labels


@pytest.fixture
def mock_quantum_processor():
    """Create a mock quantum processor"""
    processor = Mock(spec=QuantumProcessor)
    rng = np.random.default_rng(seed=42)
    processor.process_features = Mock(return_value=rng.standard_normal((100, 3)))
    return processor


@pytest.fixture
def hybrid_model(mock_quantum_processor):
    """Create a hybrid model instance with mock quantum processor"""
    config = HybridConfig(
        n_estimators=100,
        learning_rate=0.1,
        max_depth=3,
        quantum_feature_ratio=0.3,
        n_qubits=2,
        n_layers=1,
    )
    return XGBoostQuantumHybrid(config=config, quantum_processor=mock_quantum_processor)


@pytest.mark.asyncio
async def test_initialization(hybrid_model):
    """Test model initialization"""
    assert hybrid_model.config is not None
    assert hybrid_model.quantum_processor is not None
    assert hybrid_model.xgb_trainer is not None
    assert hybrid_model.enhanced_xgb is not None
    assert hybrid_model.quantum_circuit is not None
    assert hybrid_model.scaler is not None


@pytest.mark.asyncio
async def test_preprocess_features(hybrid_model, sample_data):
    """Test feature preprocessing"""
    features, labels = sample_data

    # Test without feature importance
    features_processed, labels_processed = await hybrid_model.preprocess_features(
        features, labels
    )
    assert features_processed.shape[0] == features.shape[0]
    assert labels_processed is not None

    # Test with feature importance
    rng = np.random.default_rng(seed=42)
    hybrid_model.feature_importance = rng.random(features.shape[1])
    features_processed, labels_processed = await hybrid_model.preprocess_features(
        features, labels
    )
    assert features_processed.shape[0] == features.shape[0]
    assert labels_processed is not None


@pytest.mark.asyncio
async def test_train(hybrid_model, sample_data):
    """Test model training"""
    features, labels = sample_data

    # Mock enhanced XGBoost fit method
    async def mock_fit(*args, **kwargs):
        return {"accuracy": 0.85, "roc_auc": 0.9}

    hybrid_model.enhanced_xgb.fit = mock_fit

    # Train model
    metrics = await hybrid_model.train(features, labels, validation_split=0.2)

    assert isinstance(metrics, dict)
    assert "accuracy" in metrics
    assert "roc_auc" in metrics


@pytest.mark.asyncio
async def test_predict(hybrid_model, sample_data):
    """Test model prediction"""
    features, _ = sample_data

    # Mock enhanced XGBoost predict methods
    rng = np.random.default_rng(seed=42)
    hybrid_model.enhanced_xgb.predict = Mock(
        return_value=rng.integers(0, 2, features.shape[0])
    )
    hybrid_model.enhanced_xgb.predict_proba = Mock(
        return_value=rng.random((features.shape[0], 2))
    )

    # Test regular predictions
    predictions = await hybrid_model.predict(features, return_proba=False)
    assert predictions.shape[0] == features.shape[0]
    assert np.all((predictions == 0) | (predictions == 1))

    # Test probability predictions
    predictions_proba = await hybrid_model.predict(features, return_proba=True)
    assert predictions_proba.shape[0] == features.shape[0]
    assert np.all((predictions_proba >= 0) & (predictions_proba <= 1))


@pytest.mark.asyncio
async def test_optimize_hyperparameters(
    hybrid_model: HybridModel,
    sample_data: Tuple[NDArray[np.float64], NDArray[np.float64]],
) -> None:
    """Test hyperparameter optimization"""
    features, labels = sample_data

    # Mock enhanced XGBoost optimize_hyperparameters method
    async def mock_optimize(*args: Any, **kwargs: Any) -> Dict[str, Any]:
        return {"n_estimators": 100, "learning_rate": 0.1, "max_depth": 3}

    hybrid_model.enhanced_xgb.optimize_hyperparameters = mock_optimize  # type: ignore

    # Optimize hyperparameters
    best_params = await hybrid_model.optimize_hyperparameters(
        features, labels, n_trials=10
    )

    assert isinstance(best_params, dict)
    assert "n_estimators" in best_params
    assert "learning_rate" in best_params
    assert "max_depth" in best_params


def test_get_feature_importance(hybrid_model: HybridModel) -> None:
    """Test feature importance retrieval"""
    # Test without feature importance
    with pytest.raises(ValueError):
        hybrid_model.get_feature_importance()

    # Test with feature importance
    hybrid_model.feature_importance = {"feature_1": 0.5, "feature_2": 0.3}
    importance = hybrid_model.get_feature_importance()
    assert isinstance(importance, dict)
    assert len(importance) == 2
    assert np.isclose(importance["feature_1"], 0.5, rtol=1e-7, atol=1e-7)


@pytest.mark.asyncio
async def test_error_handling(
    hybrid_model: HybridModel,
    sample_data: Tuple[NDArray[np.float64], NDArray[np.float64]],
) -> None:
    """Test error handling in various methods"""
    features, labels = sample_data

    # Test training error
    hybrid_model.enhanced_xgb.fit = Mock(side_effect=RuntimeError("Training error"))  # type: ignore
    with pytest.raises(RuntimeError):
        await hybrid_model.train(features, labels)

    # Test prediction error
    hybrid_model.enhanced_xgb.predict = Mock(  # type: ignore
        side_effect=RuntimeError("Prediction error")
    )
    with pytest.raises(RuntimeError):
        await hybrid_model.predict(features)

    # Test optimization error
    hybrid_model.enhanced_xgb.optimize_hyperparameters = Mock(  # type: ignore
        side_effect=RuntimeError("Optimization error")
    )
    with pytest.raises(RuntimeError):
        await hybrid_model.optimize_hyperparameters(features, labels)

    # Test invalid input error
    hybrid_model.enhanced_xgb.predict = Mock(  # type: ignore
        side_effect=ValueError("Invalid input data")
    )
    with pytest.raises(ValueError):
        await hybrid_model.predict(features)

    # Test dependency error
    hybrid_model.enhanced_xgb.fit = Mock(  # type: ignore
        side_effect=ImportError("Failed to import required dependencies")
    )
    with pytest.raises(ImportError):
        await hybrid_model.train(features, labels)


@pytest.mark.asyncio
async def test_fusion_weights(
    hybrid_model: HybridModel,
    sample_data: Tuple[NDArray[np.float64], NDArray[np.float64]],
) -> None:
    """Test fusion weights"""
    _, _ = sample_data  # Use _ for unused variables

    # Mock fusion weights
    output = Mock()
    output.fusion_weights = {"feature_1": 0.5, "feature_2": 0.5}

    # Check fusion weights using np.isclose for floating point comparison
    assert all(0 <= w <= 1 for w in output.fusion_weights.values())
    assert np.isclose(sum(output.fusion_weights.values()), 1.0, rtol=1e-6, atol=1e-6)


def generate_data(
    n_samples: int = 1000, n_features: int = 10
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Generate synthetic data for testing."""
    rng = np.random.default_rng(seed=42)
    X = rng.normal(0, 1, (n_samples, n_features))
    y = (X[:, 0] + X[:, 1] > 0).astype(int)
    return cast(NDArray[np.float64], X), cast(NDArray[np.float64], y)


def test_model_performance() -> None:
    """Test model performance metrics."""
    # Generate test data
    _, _ = generate_data()  # Use _ for unused variables

    # Create metrics dictionary
    metrics = {"accuracy": 0.85, "precision": 0.82, "recall": 0.88, "f1": 0.85}

    # Use np.isclose for floating point comparison
    assert np.isclose(metrics["accuracy"], 0.85, rtol=1e-5, atol=1e-5)


def test_feature_importance() -> None:
    """Test feature importance calculation."""
    # Generate test data
    _, _ = generate_data()  # Use _ for unused variables

    # ... existing code ...



"""Email service implementation."""

import logging
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Any, Dict, List, Optional

from fastapi import BackgroundTasks
from pydantic import EmailStr

from ..config import settings
from ..models.email import EmailMessage

logger = logging.getLogger(__name__)


class EmailService:
    def __init__(self):
        self.smtp_server = settings.SMTP_HOST
        self.smtp_port = settings.SMTP_PORT
        self.smtp_username = settings.SMTP_USER
        self.smtp_password = settings.SMTP_PASSWORD
        self.from_email = settings.FROM_EMAIL

    async def send_welcome_email(
        self, email: str, api_key: str, plan: str, documentation_url: str
    ) -> None:
        """Send welcome email to new subscribers."""
        try:
            subject = f"Welcome to Bleu.js {plan.title()} Plan!"

            # Create HTML content
            html_content = self._create_welcome_email_content(
                plan=plan, api_key=api_key, documentation_url=documentation_url
            )

            # Create plain text content
            text_content = self._create_welcome_email_text_content(
                plan=plan, api_key=api_key, documentation_url=documentation_url
            )

            # Create message
            msg = MIMEMultipart("alternative")
            msg["Subject"] = subject
            msg["From"] = self.from_email
            msg["To"] = email

            # Attach both HTML and plain text versions
            msg.attach(MIMEText(text_content, "plain"))
            msg.attach(MIMEText(html_content, "html"))

            # Send email
            with smtplib.SMTP(self.smtp_server, self.smtp_port) as server:
                server.starttls()
                server.login(self.smtp_username, self.smtp_password)
                server.send_message(msg)

            logger.info(f"Welcome email sent to {email}")

        except Exception as e:
            logger.error(f"Error sending welcome email: {str(e)}")
            raise

    async def send_usage_alert(self, email: str, usage_percentage: float) -> None:
        """Send usage alert email when API calls are running low."""
        try:
            subject = "Bleu.js API Usage Alert"

            # Create HTML content
            html_content = self._create_usage_alert_content(usage_percentage)

            # Create plain text content
            text_content = self._create_usage_alert_text_content(usage_percentage)

            # Create message
            msg = MIMEMultipart("alternative")
            msg["Subject"] = subject
            msg["From"] = self.from_email
            msg["To"] = email

            # Attach both HTML and plain text versions
            msg.attach(MIMEText(text_content, "plain"))
            msg.attach(MIMEText(html_content, "html"))

            # Send email
            with smtplib.SMTP(self.smtp_server, self.smtp_port) as server:
                server.starttls()
                server.login(self.smtp_username, self.smtp_password)
                server.send_message(msg)

            logger.info(f"Usage alert email sent to {email}")

        except Exception as e:
            logger.error(f"Error sending usage alert email: {str(e)}")
            raise

    def _create_welcome_email_content(
        self, plan: str, api_key: str, documentation_url: str
    ) -> str:
        """Create HTML content for welcome email."""
        return f"""
        <html>
            <head>
                <style>
                    body {{ font-family: Arial, sans-serif; line-height: 1.6; color: #333; }}
                    .container {{ max-width: 600px; margin: 0 auto; padding: 20px; }}
                    .header {{ background-color: #1a56db; color: white; padding: 20px; text-align: center; }}
                    .content {{ padding: 20px; }}
                    .api-key {{ background-color: #f3f4f6; padding: 15px; border-radius: 5px; margin: 20px 0; }}
                    .button {{ display: inline-block; padding: 10px 20px; background-color: #1a56db; color: white; text-decoration: none; border-radius: 5px; }}
                </style>
            </head>
            <body>
                <div class="container">
                    <div class="header">
                        <h1>Welcome to Bleu.js {plan.title()} Plan!</h1>
                    </div>
                    <div class="content">
                        <p>Thank you for choosing Bleu.js! We're excited to have you on board.</p>
                        <p>Your account has been set up successfully with the following details:</p>
                        <ul>
                            <li>Plan: {plan.title()}</li>
                            <li>API Key: <div class="api-key">{api_key}</div></li>
                        </ul>
                        <p>To get started, please visit our documentation:</p>
                        <p><a href="{documentation_url}" class="button">View Documentation</a></p>
                        <p>If you have any questions or need assistance, our support team is here to help.</p>
                        <p>Best regards,<br>The Bleu.js Team</p>
                    </div>
                </div>
            </body>
        </html>
        """

    def _create_welcome_email_text_content(
        self, plan: str, api_key: str, documentation_url: str
    ) -> str:
        """Create plain text content for welcome email."""
        return f"""
        Welcome to Bleu.js {plan.title()} Plan!

        Thank you for choosing Bleu.js! We're excited to have you on board.

        Your account has been set up successfully with the following details:
        - Plan: {plan.title()}
        - API Key: {api_key}

        To get started, please visit our documentation:
        {documentation_url}

        If you have any questions or need assistance, our support team is here to help.

        Best regards,
        The Bleu.js Team
        """

    def _create_usage_alert_content(self, usage_percentage: float) -> str:
        """Create HTML content for usage alert email."""
        return f"""
        <html>
            <head>
                <style>
                    body {{ font-family: Arial, sans-serif; line-height: 1.6; color: #333; }}
                    .container {{ max-width: 600px; margin: 0 auto; padding: 20px; }}
                    .header {{ background-color: #dc2626; color: white; padding: 20px; text-align: center; }}
                    .content {{ padding: 20px; }}
                    .button {{ display: inline-block; padding: 10px 20px; background-color: #1a56db; color: white; text-decoration: none; border-radius: 5px; }}
                </style>
            </head>
            <body>
                <div class="container">
                    <div class="header">
                        <h1>API Usage Alert</h1>
                    </div>
                    <div class="content">
                        <p>Your Bleu.js API usage has reached {usage_percentage}% of your monthly limit.</p>
                        <p>To ensure uninterrupted service, please consider:</p>
                        <ul>
                            <li>Upgrading your plan for higher limits</li>
                            <li>Optimizing your API usage</li>
                            <li>Contacting our support team for assistance</li>
                        </ul>
                        <p><a href="https://bleujs.org/pricing" class="button">View Plans</a></p>
                        <p>Best regards,<br>The Bleu.js Team</p>
                    </div>
                </div>
            </body>
        </html>
        """

    def _create_usage_alert_text_content(self, usage_percentage: float) -> str:
        """Create plain text content for usage alert email."""
        return f"""
        API Usage Alert

        Your Bleu.js API usage has reached {usage_percentage}% of your monthly limit.

        To ensure uninterrupted service, please consider:
        - Upgrading your plan for higher limits
        - Optimizing your API usage
        - Contacting our support team for assistance

        View our plans: https://bleujs.org/pricing

        Best regards,
        The Bleu.js Team
        """

    def send_email(
        self,
        recipient: str,
        subject: str,
        body: str,
        html_body: Optional[str] = None,
        attachments: Optional[List[Dict[str, Any]]] = None,
    ) -> bool:
        """
        Send an email to the specified recipient.

        Args:
            recipient: Email address of the recipient
            subject: Email subject
            body: Plain text email body
            html_body: Optional HTML email body
            attachments: Optional list of attachment dictionaries

        Returns:
            bool: True if email was sent successfully, False otherwise
        """
        try:
            # Create message
            msg = MIMEMultipart("alternative")
            msg["Subject"] = subject
            msg["From"] = self.from_email
            msg["To"] = recipient

            # Attach plain text content
            msg.attach(MIMEText(body, "plain"))

            # Attach HTML content if provided
            if html_body:
                msg.attach(MIMEText(html_body, "html"))

            # Send email
            with smtplib.SMTP(self.smtp_server, self.smtp_port) as server:
                server.starttls()
                server.login(self.smtp_username, self.smtp_password)
                server.send_message(msg)

            logger.info(f"Email sent to {recipient}")
            return True

        except Exception as e:
            logger.error(f"Error sending email: {str(e)}")
            return False



#!/usr/bin/env python3
"""
Bleu.js CLI - Command-line interface for the quantum-enhanced vision system
"""

import argparse
import logging
from pathlib import Path

from .quantum_detector import AdvancedQuantumDetector
from .utils import setup_logging


def main():
    """Main entry point for the CLI."""
    parser = argparse.ArgumentParser(
        description="Bleu.js - Quantum-Enhanced Vision System"
    )
    parser.add_argument("--input", "-i", required=True, help="Input image path")
    parser.add_argument("--output", "-o", help="Output directory for results")
    parser.add_argument(
        "--confidence", "-c", type=float, default=0.1, help="Confidence threshold"
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Enable verbose logging"
    )

    args = parser.parse_args()

    # Setup logging
    log_level = logging.DEBUG if args.verbose else logging.INFO
    setup_logging(log_level)

    # Initialize detector
    detector = AdvancedQuantumDetector(confidence_threshold=args.confidence)

    # Process image
    input_path = Path(args.input)
    if not input_path.exists():
        logging.error(f"Input file not found: {input_path}")
        return 1

    try:
        results = detector.detect_objects(str(input_path))
        logging.info(f"Detection results: {results}")

        # Save results if output directory specified
        if args.output:
            output_dir = Path(args.output)
            output_dir.mkdir(parents=True, exist_ok=True)
            results_file = output_dir / f"{input_path.stem}_results.json"
            results.save_to_json(str(results_file))
            logging.info(f"Results saved to: {results_file}")

        return 0
    except Exception as e:
        logging.error(f"Error processing image: {e}")
        return 1


if __name__ == "__main__":
    exit(main())



import asyncio
import secrets
import uuid
from datetime import datetime, timedelta, timezone

import pytest
from fastapi import HTTPException
from fastapi.testclient import TestClient
from sqlalchemy.orm import Session

from src.database import get_db
from src.main import app
from src.models.subscription import (
    APIToken,
    APITokenCreate,
    Subscription,
    SubscriptionPlan,
)
from src.models.user import User
from src.services import init_services
from src.services.api_token_service import APITokenService
from src.services.auth_service import AuthService

client = TestClient(app)


@pytest.fixture
def services(db: Session):
    return init_services(db)


@pytest.fixture
def test_user(db: Session) -> User:
    """Create a test user with an active subscription."""
    user = User(
        id=str(uuid.uuid4()),
        email="test@example.com",
        hashed_password="hashed_password",
        full_name="Test User",
        is_active=True,
        is_verified=True,
        trial_end_date=datetime.now(timezone.utc) + timedelta(days=30),
    )
    db.add(user)
    db.commit()
    db.refresh(user)

    # Create a subscription plan first
    plan = SubscriptionPlan(
        id=str(uuid.uuid4()),
        name="Test Plan",
        plan_type="CORE",
        price=1000,
        api_calls_limit=100,
        rate_limit=1000,
        uptime_sla="99.9",
        support_level="standard",
        features={"core_ai_model_access": True},
        trial_days=30,
    )
    db.add(plan)
    db.commit()

    subscription = Subscription(
        id=str(uuid.uuid4()),
        user_id=user.id,
        plan_id=plan.id,
        plan_type="CORE",
        status="active",
        current_period_start=datetime.now(timezone.utc),
        current_period_end=datetime.now(timezone.utc) + timedelta(days=30),
    )
    db.add(subscription)
    db.commit()

    return user


@pytest.fixture
def test_subscription(db: Session, test_user: User):
    plan = SubscriptionPlan(
        id="test_plan_id",
        name="Test Plan",
        plan_type="CORE",
        price=1000,
        api_calls_limit=100,
        rate_limit=1000,
        uptime_sla=99.9,
        support_level="standard",
        features={"core_ai_model_access": True},
    )
    db.add(plan)
    db.commit()

    subscription = Subscription(
        id="test_subscription_id",
        user_id=test_user.id,
        plan_id=plan.id,
        plan_type="CORE",
        status="active",
        current_period_start=datetime.now(timezone.utc),
        current_period_end=datetime.now(timezone.utc) + timedelta(days=30),
    )
    db.add(subscription)
    db.commit()
    return subscription


@pytest.fixture
def test_token(db: Session, test_user: User) -> APIToken:
    """Create a test API token."""
    token = APIToken(
        id=str(uuid.uuid4()),
        user_id=test_user.id,
        subscription_id=test_user.subscription.id,
        name="Test Token",
        token=secrets.token_urlsafe(32),
        is_active=True,
        expires_at=datetime.now(timezone.utc) + timedelta(days=30),
    )
    db.add(token)
    db.commit()
    db.refresh(token)
    return token


@pytest.fixture
def auth_headers(test_user: User, db: Session) -> dict:
    """Create authentication headers for test user."""
    auth_service = AuthService(db)
    access_token = auth_service.create_access_token({"sub": test_user.email})
    return {"Authorization": f"Bearer {access_token}"}


async def test_create_token(client: TestClient, test_user: User, auth_headers: dict):
    """Test creating a new API token."""
    response = client.post(
        "/api/v1/tokens",
        json={"name": "New Token"},
        headers=auth_headers,
    )
    assert response.status_code == 200
    data = response.json()
    assert data["name"] == "New Token"
    assert data["user_id"] == test_user.id
    assert data["is_active"] is True


async def test_create_token_without_subscription(
    client: TestClient, db: Session, auth_headers: dict
):
    """Test creating a token without an active subscription."""
    user = User(
        id=str(uuid.uuid4()),
        email="no_sub@example.com",
        hashed_password="hashed_password",
        full_name="No Subscription User",
        is_active=True,
        is_verified=True,
        trial_end_date=datetime.now(timezone.utc) + timedelta(days=30),
    )
    db.add(user)
    db.commit()

    auth_service = AuthService(db)
    access_token = auth_service.create_access_token({"sub": user.email})
    headers = {"Authorization": f"Bearer {access_token}"}

    response = client.post(
        "/api/v1/tokens",
        json={"name": "New Token"},
        headers=headers,
    )
    assert response.status_code == 403


async def test_get_tokens(
    client: TestClient, test_user: User, test_token: APIToken, auth_headers: dict
):
    """Test getting all user's API tokens."""
    response = client.get("/api/v1/tokens", headers=auth_headers)
    assert response.status_code == 200
    data = response.json()
    assert len(data) == 1
    assert data[0]["id"] == test_token.id


async def test_revoke_token(
    client: TestClient, test_user: User, test_token: APIToken, auth_headers: dict
):
    """Test revoking an API token."""
    response = client.post(
        f"/api/v1/tokens/{test_token.id}/revoke",
        headers=auth_headers,
    )
    assert response.status_code == 200
    data = response.json()
    assert data["id"] == test_token.id
    assert data["is_active"] is False


async def test_rotate_token(
    client: TestClient, test_user: User, test_token: APIToken, auth_headers: dict
):
    """Test rotating an API token."""
    old_token = test_token.token
    response = client.post(
        f"/api/v1/tokens/{test_token.id}/rotate",
        headers=auth_headers,
    )
    assert response.status_code == 200
    data = response.json()
    assert data["id"] == test_token.id
    assert data["token"] != old_token


async def test_validate_token(db: Session, test_token: APIToken):
    """Test validating an API token."""
    token_service = APITokenService(db)
    result = await token_service.validate_token(test_token.token)
    assert result is True

    # Test invalid token
    result = await token_service.validate_token("invalid_token")
    assert result is False


async def test_validate_expired_token(db: Session):
    """Test validating an expired token."""
    # Create a test user with subscription
    user = User(
        id=str(uuid.uuid4()),
        email="test@example.com",
        hashed_password="hashed_password",
        is_active=True,
        is_verified=True,
    )
    db.add(user)
    db.commit()

    # Create a subscription plan
    plan = SubscriptionPlan(
        id=str(uuid.uuid4()),
        name="Test Plan",
        plan_type="CORE",
        price=1000,
        api_calls_limit=100,
        rate_limit=1000,
        uptime_sla="99.9",
        support_level="standard",
        features={"core_ai_model_access": True},
        trial_days=30,
    )
    db.add(plan)
    db.commit()

    # Create a subscription
    subscription = Subscription(
        id=str(uuid.uuid4()),
        user_id=user.id,
        plan_id=plan.id,
        plan_type="CORE",
        status="active",
        current_period_start=datetime.now(timezone.utc),
        current_period_end=datetime.now(timezone.utc) + timedelta(days=30),
    )
    db.add(subscription)
    db.commit()

    # Create an expired token
    token = APIToken(
        id=str(uuid.uuid4()),
        user_id=user.id,
        subscription_id=subscription.id,
        name="Expired Token",
        token=secrets.token_urlsafe(32),
        is_active=True,
        expires_at=datetime.now(timezone.utc) - timedelta(days=1),
    )
    db.add(token)
    db.commit()

    # Test validation
    token_service = APITokenService(db)
    result = await token_service.validate_token(token.token)
    assert result is False


@pytest.mark.asyncio
async def test_create_api_token(services: dict, test_user: User):
    """Test creating an API token using the service directly."""
    token_data = APITokenCreate(name="Service Token")
    token = await services["api_token_service"].create_token(test_user, token_data)
    assert token.name == "Service Token"
    assert token.user_id == test_user.id
    assert token.is_active is True


@pytest.mark.asyncio
async def test_create_api_token_without_subscription(services: dict, db: Session):
    """Test creating a token without subscription using the service directly."""
    user = User(
        id=str(uuid.uuid4()),
        email="no_sub_service@example.com",
        hashed_password="hashed_password",
        is_active=True,
        is_verified=True,
    )
    db.add(user)
    db.commit()

    token_data = APITokenCreate(name="No Sub Token")
    with pytest.raises(HTTPException) as exc_info:
        await services["api_token_service"].create_token(user, token_data)
    assert exc_info.value.status_code == 403


@pytest.mark.asyncio
async def test_get_user_tokens(services: dict, test_user: User, test_token: APIToken):
    """Test getting user tokens using the service directly."""
    tokens = await services["api_token_service"].get_user_tokens(test_user)
    assert len(tokens) == 1
    assert tokens[0].id == test_token.id


@pytest.mark.asyncio
async def test_revoke_api_token(services: dict, test_user: User, test_token: APIToken):
    """Test revoking a token using the service directly."""
    result = await services["api_token_service"].revoke_token(test_token.id, test_user)
    assert result.id == test_token.id
    assert result.is_active is False


@pytest.mark.asyncio
async def test_revoke_nonexistent_token(services: dict, test_user: User):
    """Test revoking a non-existent token using the service directly."""
    with pytest.raises(HTTPException) as exc_info:
        await services["api_token_service"].revoke_token("nonexistent", test_user)
    assert exc_info.value.status_code == 404


@pytest.mark.asyncio
async def test_rotate_api_token(services: dict, test_user: User, test_token: APIToken):
    """Test rotating a token using the service directly."""
    old_token = test_token.token
    new_token = await services["api_token_service"].rotate_token(
        test_token.id, test_user
    )
    assert new_token.id == test_token.id
    assert new_token.token != old_token


@pytest.mark.asyncio
async def test_create_multiple_tokens(services: dict, test_user: User):
    """Test creating multiple tokens for the same user."""
    token_data1 = APITokenCreate(name="Token 1")
    token_data2 = APITokenCreate(name="Token 2")

    token1 = await services["api_token_service"].create_token(test_user, token_data1)
    token2 = await services["api_token_service"].create_token(test_user, token_data2)

    assert token1.name == "Token 1"
    assert token2.name == "Token 2"
    assert token1.token != token2.token


@pytest.mark.asyncio
async def test_token_expiration(services: dict, test_user: User):
    """Test token expiration."""
    token_data = APITokenCreate(
        name="Expiring Token",
        expires_at=datetime.now(timezone.utc) + timedelta(days=7),
    )

    token = await services["api_token_service"].create_token(test_user, token_data)
    assert token.name == "Expiring Token"
    assert token.expires_at is not None



"""Tests for quantum circuit implementation."""

import numpy as np
import pytest
from qiskit import QuantumCircuit
from qiskit.quantum_info import Operator, Statevector

from src.quantum.core.quantum_circuit import QuantumCircuit as CustomQuantumCircuit


@pytest.mark.quantum
@pytest.mark.quantum_simulator
class TestQuantumCircuit:
    """Test suite for quantum circuit implementation."""

    @pytest.fixture
    def custom_circuit(self):
        """Create a custom quantum circuit instance."""
        return CustomQuantumCircuit(num_qubits=2)

    def test_initialization(self, custom_circuit):
        """Test circuit initialization."""
        assert custom_circuit.num_qubits == 2
        assert custom_circuit.depth == 0
        assert len(custom_circuit.gates) == 0

    def test_add_custom_gate(self, custom_circuit):
        """Test adding custom gates."""
        # Hadamard matrix
        h_matrix = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
        custom_circuit.add_custom_gate(
            name="custom_h",
            matrix=h_matrix,
            target_qubits=[0]
        )
        assert len(custom_circuit.gates) == 1
        assert custom_circuit.depth == 1

    def test_gate_merging(self, custom_circuit):
        """Test gate merging functionality."""
        # Add two compatible gates
        h_matrix = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
        custom_circuit.add_custom_gate(
            name="h",
            matrix=h_matrix,
            target_qubits=[0]
        )
        custom_circuit.add_custom_gate(
            name="h",
            matrix=h_matrix,
            target_qubits=[1]
        )
        # Gates should be merged into same layer
        assert custom_circuit.depth == 1

    def test_measurement_statistics(self, custom_circuit, quantum_simulator):
        """Test measurement statistics calculation."""
        # Prepare |+⟩ state
        h_matrix = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
        custom_circuit.add_custom_gate(
            name="h",
            matrix=h_matrix,
            target_qubits=[0]
        )
        
        stats = custom_circuit.get_measurement_statistics(qubit_index=0)
        assert len(stats) == 2
        np.testing.assert_almost_equal(stats[0], 0.5, decimal=2)
        np.testing.assert_almost_equal(stats[1], 0.5, decimal=2)

    def test_circuit_validation(self, custom_circuit):
        """Test circuit validation checks."""
        # Invalid qubit index
        with pytest.raises(ValueError):
            custom_circuit.add_custom_gate(
                name="test",
                matrix=np.eye(2),
                target_qubits=[5]  # Invalid qubit index
            )

        # Non-unitary matrix
        non_unitary = np.array([[1, 1], [1, 1]])
        with pytest.raises(ValueError):
            custom_circuit.add_custom_gate(
                name="invalid",
                matrix=non_unitary,
                target_qubits=[0]
            )

    def test_quantum_state_evolution(self, custom_circuit, quantum_simulator):
        """Test quantum state evolution through gates."""
        # Create Bell state
        h_matrix = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
        cnot_matrix = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 0, 1],
            [0, 0, 1, 0]
        ])
        
        custom_circuit.add_custom_gate(
            name="h",
            matrix=h_matrix,
            target_qubits=[0]
        )
        custom_circuit.add_custom_gate(
            name="cnot",
            matrix=cnot_matrix,
            target_qubits=[0, 1]
        )
        
        # Verify Bell state
        expected_state = np.array([1, 0, 0, 1]) / np.sqrt(2)
        final_state = custom_circuit.get_statevector()
        np.testing.assert_array_almost_equal(final_state, expected_state)

    @pytest.mark.quantum_benchmark
    def test_circuit_performance(self, custom_circuit, quantum_test_utils):
        """Test circuit performance metrics."""
        random_circuit = quantum_test_utils["create_random_circuit"](
            num_qubits=2,
            depth=10,
            seed=42
        )
        
        # Convert random circuit to custom circuit
        for gate in random_circuit.data:
            if gate[0].name == "h":
                h_matrix = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
                custom_circuit.add_custom_gate(
                    name="h",
                    matrix=h_matrix,
                    target_qubits=[gate[1][0].index]
                )
            elif gate[0].name == "cx":
                cnot_matrix = np.array([
                    [1, 0, 0, 0],
                    [0, 1, 0, 0],
                    [0, 0, 0, 1],
                    [0, 0, 1, 0]
                ])
                custom_circuit.add_custom_gate(
                    name="cnot",
                    matrix=cnot_matrix,
                    target_qubits=[gate[1][0].index, gate[1][1].index]
                )
        
        # Verify circuit metrics
        assert custom_circuit.depth <= 10  # Due to potential gate merging
        assert len(custom_circuit.gates) > 0 


import uuid
from datetime import datetime, timezone
from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, EmailStr, Field
from sqlalchemy import JSON, Boolean, Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from src.models.declarative_base import Base


class Customer(Base):
    """Database model for storing customer information."""

    __tablename__ = "customers"
    __table_args__ = {"extend_existing": True}

    id = Column(String, primary_key=True, default=lambda: str(uuid.uuid4()))
    stripe_customer_id = Column(String, unique=True, nullable=False)
    email = Column(String, unique=True, nullable=False)
    api_key = Column(String, unique=True, nullable=False)
    plan = Column(String, nullable=False)
    features = Column(JSON, nullable=False)
    api_calls_remaining = Column(Integer, default=0)
    rate_limit = Column(Integer, nullable=False)
    subscription_start = Column(DateTime, nullable=False)
    subscription_end = Column(DateTime, nullable=True)
    is_active = Column(Boolean, default=True)
    created_at = Column(DateTime, default=lambda: datetime.now(timezone.utc))
    updated_at = Column(
        DateTime,
        default=lambda: datetime.now(timezone.utc),
        onupdate=lambda: datetime.now(timezone.utc),
    )
    subscription_id = Column(String, ForeignKey("subscriptions.id"), nullable=True)
    api_calls_reset_at = Column(DateTime, nullable=True)
    settings = Column(JSON, default={})

    # Relationships
    api_calls = relationship("APICall", back_populates="customer")
    rate_limit_tokens = relationship(
        "RateLimitToken", back_populates="customer", uselist=False
    )
    subscription = relationship("Subscription", back_populates="customer")
    api_tokens = relationship("APIToken", back_populates="customer")
    rate_limits = relationship("RateLimit", back_populates="customer")
    payments = relationship("Payment", back_populates="customer")

    def reset_api_calls(self):
        """Reset the API calls counter."""
        self.api_calls_remaining = self.subscription.plan.api_calls_limit
        self.api_calls_reset_at = datetime.now(timezone.utc)

    def can_make_api_call(self) -> bool:
        """Check if the customer can make an API call."""
        if not self.api_calls_reset_at:
            return False

        # Reset if the period has elapsed
        now = datetime.now(timezone.utc)
        if (now - self.api_calls_reset_at).days >= 30:  # Monthly reset
            self.reset_api_calls()

        return self.api_calls_remaining > 0

    def decrement_api_calls(self):
        """Decrement the API calls counter."""
        if self.can_make_api_call():
            self.api_calls_remaining -= 1
            self.updated_at = datetime.now(timezone.utc)
            return True
        return False


class RateLimitToken(Base):
    """Database model for managing rate limiting tokens."""

    __tablename__ = "rate_limit_tokens"
    __table_args__ = {"extend_existing": True}

    id = Column(String, primary_key=True)
    customer_id = Column(
        String, ForeignKey("customers.id"), unique=True, nullable=False
    )
    tokens = Column(Integer, nullable=False)
    last_updated = Column(DateTime, nullable=False)

    # Relationships
    customer = relationship("Customer", back_populates="rate_limit_tokens")


# Pydantic models for API
class CustomerBase(BaseModel):
    email: EmailStr
    plan: str
    features: Dict
    api_calls_remaining: int
    rate_limit: int

    model_config = ConfigDict(arbitrary_types_allowed=True)


class CustomerCreate(CustomerBase):
    stripe_customer_id: str
    api_key: str
    subscription_start: datetime

    model_config = ConfigDict(arbitrary_types_allowed=True)


class CustomerUpdate(BaseModel):
    plan: Optional[str] = None
    features: Optional[Dict] = None
    api_calls_remaining: Optional[int] = None
    rate_limit: Optional[int] = None
    is_active: Optional[bool] = None
    subscription_end: Optional[datetime] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class CustomerResponse(CustomerBase):
    id: str
    stripe_customer_id: str
    api_key: str
    subscription_start: datetime
    subscription_end: Optional[datetime]
    is_active: bool
    created_at: datetime
    updated_at: datetime

    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True)



import os
from datetime import datetime, timezone

import pytest
from dotenv import load_dotenv
from fastapi.testclient import TestClient
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import StaticPool

# Set testing environment and load test environment variables
os.environ["TESTING"] = "true"
load_dotenv(".env.test")

from src.main import app
from src.models.customer import Customer
from src.models.declarative_base import Base

# Import test settings and models after setting TESTING=true
from tests.test_config import get_test_settings

# Get test settings
settings = get_test_settings()

# Create test database engine - always use SQLite for testing
engine = create_engine(
    "sqlite:///:memory:",
    connect_args={"check_same_thread": False},
    poolclass=StaticPool,
)

# Create test session factory
TestingSessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


@pytest.fixture(scope="function")
def db_engine():
    """Create and drop database schema for each test."""
    Base.metadata.create_all(bind=engine)
    yield engine
    Base.metadata.drop_all(bind=engine)


@pytest.fixture(scope="function")
def db(db_engine):
    """Get a database session for testing."""
    connection = db_engine.connect()
    transaction = connection.begin()
    session = TestingSessionLocal(bind=connection)

    yield session

    session.close()
    transaction.rollback()
    connection.close()


@pytest.fixture(scope="function")
def client(db):
    """Get a test client with database session."""
    from src.database import get_db

    def override_get_db():
        try:
            yield db
        finally:
            pass  # Don't close the session here as it's managed by the db fixture

    app.dependency_overrides[get_db] = override_get_db
    with TestClient(app) as test_client:
        yield test_client
    app.dependency_overrides.clear()


@pytest.fixture(scope="function")
def clean_db(db):
    # Clean up all tables
    for table in reversed(Base.metadata.sorted_tables):
        db.execute(table.delete())
    db.commit()


@pytest.fixture
def test_customer(db):
    """Create a test customer."""
    customer = Customer(
        stripe_customer_id="test_stripe_customer_id",
        email="test@example.com",
        api_key="test_api_key",
        plan="core",
        features=["api_access"],
        rate_limit=100,
        subscription_start=datetime.now(timezone.utc),
        subscription_end=datetime.now(timezone.utc).replace(
            year=datetime.now(timezone.utc).year + 1
        ),
        is_active=True,
    )
    db.add(customer)
    db.commit()
    return customer



"""Rate limiting configuration."""

from pydantic import BaseModel, ConfigDict


class RateLimitingConfig(BaseModel):
    """Rate limiting configuration."""

    enabled: bool = True
    rate_limit: int = 100  # requests per second
    burst_limit: int = 200  # maximum burst size
    window_size: int = 60  # window size in seconds
    key_prefix: str = "rate_limit:"  # Redis key prefix
    algorithm: str = "fixed_window"  # Rate limiting algorithm to use
    error_code: int = 429  # HTTP status code for rate limit exceeded
    error_message: str = "Rate limit exceeded. Please try again later."

    model_config = ConfigDict(arbitrary_types_allowed=True)



"""Security headers configuration."""

from pydantic import BaseModel, ConfigDict


class SecurityHeadersConfig(BaseModel):
    """Security headers configuration."""

    x_frame_options: str = "DENY"
    x_content_type_options: str = "nosniff"
    x_xss_protection: str = "1; mode=block"
    strict_transport_security: str = "max-age=31536000; includeSubDomains"
    content_security_policy: str = (
        "default-src 'self'; "
        "script-src 'self' 'unsafe-inline' 'unsafe-eval'; "
        "style-src 'self' 'unsafe-inline'; "
        "img-src 'self' data:; "
        "font-src 'self'; "
        "connect-src 'self'"
    )
    referrer_policy: str = "strict-origin-when-cross-origin"
    permissions_policy: str = (
        "accelerometer=(), "
        "camera=(), "
        "geolocation=(), "
        "gyroscope=(), "
        "magnetometer=(), "
        "microphone=(), "
        "payment=(), "
        "usb=()"
    )
    cross_origin_opener_policy: str = "same-origin"
    cross_origin_embedder_policy: str = "require-corp"
    cross_origin_resource_policy: str = "same-origin"

    model_config = ConfigDict(arbitrary_types_allowed=True)



"""
Ensemble Model Implementation
Combines multiple models for improved predictions.
"""

import logging
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from sklearn.base import BaseEstimator


class EnsembleModel:
    def __init__(
        self,
        models: Optional[List[BaseEstimator]] = None,
        weights: Optional[List[float]] = None,
    ):
        self.models = models or []
        self.weights = weights or []
        self.metrics = {}

    def add_model(self, model: BaseEstimator, weight: float = 1.0):
        """Add a model to the ensemble."""
        self.models.append(model)
        self.weights.append(weight)
        # Normalize weights
        self.weights = [w / sum(self.weights) for w in self.weights]

    def predict(
        self, X: np.ndarray, return_proba: bool = False
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Make predictions using the ensemble of models."""
        try:
            predictions = []
            probabilities = []

            for model in self.models:
                if return_proba:
                    pred, proba = model.predict(X, return_proba=True)
                    predictions.append(pred)
                    probabilities.append(proba)
                else:
                    predictions.append(model.predict(X))

            # Weighted voting for classification
            if return_proba:
                weighted_proba = np.zeros_like(probabilities[0])
                for proba, weight in zip(probabilities, self.weights):
                    weighted_proba += proba * weight
                final_pred = np.argmax(weighted_proba, axis=1)
                return final_pred, weighted_proba
            else:
                weighted_pred = np.zeros_like(predictions[0])
                for pred, weight in zip(predictions, self.weights):
                    weighted_pred += pred * weight
                return np.round(weighted_pred).astype(int)

        except Exception as e:
            logging.error(f"❌ Ensemble prediction failed: {str(e)}")
            raise

    def get_metrics(self) -> Dict:
        """Get ensemble model performance metrics."""
        return self.metrics

    def set_weights(self, weights: List[float]):
        """Set custom weights for the ensemble models."""
        if len(weights) != len(self.models):
            raise ValueError("Number of weights must match number of models")
        self.weights = [w / sum(weights) for w in weights]  # Normalize weights



import json
import os
from datetime import datetime
from typing import Dict, List, Literal, Optional

import boto3
import requests

HealthStatus = Literal["healthy", "unhealthy"]


class InfrastructureHealthCheck:
    def __init__(self):
        self.eb = boto3.client("elasticbeanstalk", region_name="us-west-2")
        self.ec2 = boto3.client("ec2", region_name="us-west-2")
        self.elb = boto3.client("elbv2", region_name="us-west-2")
        self.environment_name = "bleujs-api-prod"
        self.results = {
            "timestamp": datetime.now().isoformat(),
            "tests": [],
            "overall_status": "pending",
        }

    def log_test(self, test_name: str, status: str, details: Dict):
        """Log test results."""
        self.results["tests"].append(
            {
                "name": test_name,
                "status": status,
                "details": details,
                "timestamp": datetime.now().isoformat(),
            }
        )

    def check_eb_health(self) -> Dict:
        """Check Elastic Beanstalk environment health."""
        try:
            response = self.eb.describe_environments(
                EnvironmentNames=[self.environment_name]
            )
            env = response["Environments"][0]
            status = {
                "health": env["Health"],
                "status": env["Status"],
                "version_label": env["VersionLabel"],
                "cname": env["CNAME"],
            }
            self.log_test("eb_health", "success", status)
            return status
        except Exception as e:
            self.log_test("eb_health", "failure", {"error": str(e)})
            return {"error": str(e)}

    def check_ec2_instances(self) -> Dict:
        """Check EC2 instances status."""
        try:
            # Get instance IDs from EB environment
            env_resources = self.eb.describe_environment_resources(
                EnvironmentName=self.environment_name
            )
            instance_ids = [
                i["Id"] for i in env_resources["EnvironmentResources"]["Instances"]
            ]

            # Check instance status
            instances = self.ec2.describe_instances(InstanceIds=instance_ids)
            instance_status = []
            for reservation in instances["Reservations"]:
                for instance in reservation["Instances"]:
                    status = {
                        "instance_id": instance["InstanceId"],
                        "state": instance["State"]["Name"],
                        "type": instance["InstanceType"],
                        "az": instance["Placement"]["AvailabilityZone"],
                    }
                    instance_status.append(status)

            self.log_test("ec2_instances", "success", {"instances": instance_status})
            return {"instances": instance_status}
        except Exception as e:
            self.log_test("ec2_instances", "failure", {"error": str(e)})
            return {"error": str(e)}

    def check_load_balancer(self) -> Dict:
        """Check Load Balancer health."""
        try:
            # Get load balancer ARN from EB environment
            env_resources = self.eb.describe_environment_resources(
                EnvironmentName=self.environment_name
            )
            lb_name = env_resources["EnvironmentResources"]["LoadBalancers"][0]["Name"]

            # Get load balancer details
            lbs = self.elb.describe_load_balancers(Names=[lb_name])
            lb = lbs["LoadBalancers"][0]

            # Get target group health
            target_groups = self.elb.describe_target_groups(
                LoadBalancerArn=lb["LoadBalancerArn"]
            )

            health_status = []
            for tg in target_groups["TargetGroups"]:
                health = self.elb.describe_target_health(
                    TargetGroupArn=tg["TargetGroupArn"]
                )
                health_status.append(
                    {
                        "target_group": tg["TargetGroupName"],
                        "targets": health["TargetHealthDescriptions"],
                    }
                )

            status = {
                "load_balancer": lb["LoadBalancerName"],
                "dns_name": lb["DNSName"],
                "state": lb["State"]["Code"],
                "target_groups": health_status,
            }

            self.log_test("load_balancer", "success", status)
            return status
        except Exception as e:
            self.log_test("load_balancer", "failure", {"error": str(e)})
            return {"error": str(e)}

    def check_api_endpoints(self) -> Dict:
        """Check API endpoints health."""
        try:
            # Get environment URL
            env = self.eb.describe_environments(
                EnvironmentNames=[self.environment_name]
            )["Environments"][0]
            base_url = f"https://{env['CNAME']}"

            endpoints = ["/health", "/v1/subscriptions/plans", "/v1/auth/validate"]

            results = {}
            for endpoint in endpoints:
                try:
                    response = requests.get(f"{base_url}{endpoint}")
                    results[endpoint] = {
                        "status_code": response.status_code,
                        "response_time": response.elapsed.total_seconds(),
                    }
                except requests.RequestException as e:
                    results[endpoint] = {"error": str(e)}

            self.log_test("api_endpoints", "success", results)
            return results
        except Exception as e:
            self.log_test("api_endpoints", "failure", {"error": str(e)})
            return {"error": str(e)}

    def generate_report(self) -> Dict:
        """Generate comprehensive health check report."""
        try:
            # Run all checks and store results
            health_status = {
                "elastic_beanstalk": self.check_eb_health(),
                "ec2_instances": self.check_ec2_instances(),
                "load_balancer": self.check_load_balancer(),
                "api_endpoints": self.check_api_endpoints(),
            }

            # Add health status to results
            self.results["health_status"] = health_status

            # Determine overall status
            all_success = all(
                test["status"] == "success" for test in self.results["tests"]
            )
            self.results["overall_status"] = "healthy" if all_success else "unhealthy"

            # Save report
            report_path = (
                f"reports/health_check_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"
            )
            os.makedirs("reports", exist_ok=True)
            with open(report_path, "w") as f:
                json.dump(self.results, f, indent=2)

            return self.results
        except Exception as e:
            return {"error": "Failed to generate report", "details": str(e)}


def check_elastic_beanstalk_health() -> HealthStatus:
    """Check Elastic Beanstalk environment health."""
    try:
        eb = boto3.client("elasticbeanstalk")
        response = eb.describe_environments(EnvironmentNames=["bleujs-api-prod"])
        if response["Environments"][0]["Health"] == "Green":
            return "healthy"
        return "unhealthy"
    except Exception as e:
        print(f"Error checking Elastic Beanstalk health: {str(e)}")
        return "unhealthy"


def check_ec2_health() -> HealthStatus:
    """Check EC2 instance health."""
    try:
        ec2 = boto3.client("ec2")
        response = ec2.describe_instance_status()
        for status in response["InstanceStatuses"]:
            if status["InstanceStatus"]["Status"] != "ok":
                return "unhealthy"
        return "healthy"
    except Exception as e:
        print(f"Error checking EC2 health: {str(e)}")
        return "unhealthy"


def check_load_balancer_health() -> HealthStatus:
    """Check Load Balancer health."""
    try:
        elb = boto3.client("elbv2")
        eb = boto3.client("elasticbeanstalk")

        # Get environment resources to find target group
        env_resources = eb.describe_environment_resources(
            EnvironmentName="bleujs-api-prod"
        )

        # Get target group ARN from environment resources
        target_groups = env_resources["EnvironmentResources"]["TargetGroups"]
        if not target_groups:
            print("No target groups found in environment")
            return "unhealthy"

        target_group_arn = target_groups[0]["Name"]

        # Now check the health
        response = elb.describe_target_health(TargetGroupArn=target_group_arn)

        for target in response["TargetHealthDescriptions"]:
            if target["TargetHealth"]["State"] != "healthy":
                return "unhealthy"
        return "healthy"
    except Exception as e:
        print(f"Error checking Load Balancer health: {str(e)}")
        return "unhealthy"


def check_api_health() -> HealthStatus:
    """Check API endpoint health."""
    try:
        response = requests.get(
            "https://bleujs-api-prod.eba-5k3sxpbd.us-west-2.elasticbeanstalk.com/health"
        )
        if response.status_code == 200:
            return "healthy"
        return "unhealthy"
    except Exception as e:
        print(f"Error checking API health: {str(e)}")
        return "unhealthy"


def run_health_checks():
    """Run all infrastructure health checks and return results."""
    try:
        # Get health check results
        eb_health = check_elastic_beanstalk_health()
        ec2_status = check_ec2_health()
        lb_status = check_load_balancer_health()
        api_status = check_api_health()

        # Print results
        print("Infrastructure Health Check Results:")
        print(f"Elastic Beanstalk Health: {eb_health}")
        print(f"EC2 Instance Status: {ec2_status}")
        print(f"Load Balancer Status: {lb_status}")
        print(f"API Endpoint Status: {api_status}")

        # Return overall health status
        health_results = {
            "eb_health": eb_health,
            "ec2_status": ec2_status,
            "lb_status": lb_status,
            "api_status": api_status,
        }
        return all(status == "healthy" for status in health_results.values())
    except Exception as e:
        print(f"Error running health checks: {str(e)}")
        return False


if __name__ == "__main__":
    run_health_checks()



import logging
import os

import uvicorn
from fastapi import FastAPI, HTTPException, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

from src.config import get_settings
from src.database import init_db
from src.quantum_py.core.quantum_processor import ProcessorConfig, QuantumProcessor
from src.routes import api_tokens, auth, subscription
from tests.test_config import get_test_settings

# Constants
API_V1_PREFIX = "/api/v1"
RENDER_ERROR_MSG = "Error rendering page"

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Quantum configuration
QUANTUM_CONFIG = ProcessorConfig(
    num_qubits=5,
    error_rate=0.001,
    decoherence_time=1000.0,
    gate_time=0.1,
    num_workers=4,
    max_depth=1000,
    optimization_level=1,
    use_error_correction=True,
    noise_model="depolarizing",
)

app = FastAPI(
    title="Bleu.js API",
    description=(
        "A state-of-the-art quantum-enhanced vision system "
        "with advanced AI capabilities"
    ),
    version="1.1.4",
)

# Get settings based on environment
settings = get_test_settings() if os.getenv("TESTING") == "true" else get_settings()

# Configure CORS
app.add_middleware(
    CORSMiddleware,
    allow_origins=[settings.CORS_ORIGINS],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


# Error handling middleware
@app.middleware("http")
async def error_handling_middleware(request: Request, call_next):
    try:
        response = await call_next(request)
        return response
    except Exception as e:
        logger.error(f"Error processing request: {str(e)}", exc_info=True)
        return JSONResponse(
            status_code=500,
            content={"detail": "An internal server error occurred"},
        )


# Mount static files
app.mount("/static", StaticFiles(directory="src/static"), name="static")

# Templates
templates = Jinja2Templates(directory="src/templates")

# Initialize database
if not os.getenv("TESTING"):
    try:
        init_db()
    except Exception as e:
        logger.error(f"Database initialization failed: {str(e)}", exc_info=True)
        raise HTTPException(status_code=500, detail="Database initialization failed")

# Include routers
app.include_router(auth.router, prefix=API_V1_PREFIX, tags=["auth"])
app.include_router(subscription.router, prefix=API_V1_PREFIX, tags=["subscription"])
app.include_router(api_tokens.router, prefix=API_V1_PREFIX, tags=["api_tokens"])


# Serve HTML pages
@app.get("/", response_class=HTMLResponse)
async def home(request: Request):
    try:
        return templates.TemplateResponse("index.html", {"request": request})
    except Exception as e:
        logger.error(f"Error rendering home page: {str(e)}", exc_info=True)
        raise HTTPException(status_code=500, detail=RENDER_ERROR_MSG)


@app.get("/signup", response_class=HTMLResponse)
async def signup_page(request: Request):
    try:
        return templates.TemplateResponse("signup.html", {"request": request})
    except Exception as e:
        logger.error(f"Error rendering signup page: {str(e)}", exc_info=True)
        raise HTTPException(status_code=500, detail=RENDER_ERROR_MSG)


@app.get("/signin", response_class=HTMLResponse)
async def signin_page(request: Request):
    try:
        return templates.TemplateResponse("signin.html", {"request": request})
    except Exception as e:
        logger.error(f"Error rendering signin page: {str(e)}", exc_info=True)
        raise HTTPException(status_code=500, detail=RENDER_ERROR_MSG)


@app.get("/forgot-password", response_class=HTMLResponse)
async def forgot_password_page(request: Request):
    try:
        return templates.TemplateResponse("forgot_password.html", {"request": request})
    except Exception as e:
        logger.error(f"Error rendering forgot password page: {str(e)}", exc_info=True)
        raise HTTPException(status_code=500, detail=RENDER_ERROR_MSG)


@app.get("/dashboard", response_class=HTMLResponse)
async def dashboard_page(request: Request):
    try:
        return templates.TemplateResponse("dashboard.html", {"request": request})
    except Exception as e:
        logger.error(f"Error rendering dashboard page: {str(e)}", exc_info=True)
        raise HTTPException(status_code=500, detail=RENDER_ERROR_MSG)


@app.get("/subscription", response_class=HTMLResponse)
async def subscription_dashboard(request: Request):
    try:
        return templates.TemplateResponse(
            "subscription_dashboard.html", {"request": request}
        )
    except Exception as e:
        logger.error(f"Error rendering subscription dashboard: {str(e)}", exc_info=True)
        raise HTTPException(status_code=500, detail=RENDER_ERROR_MSG)


@app.get("/")
async def root():
    return {
        "message": "Welcome to Bleu.js API",
        "version": "1.1.4",
        "documentation": "/docs",
    }


def initialize_quantum_system():
    """Initialize the quantum computing system with necessary configurations and validations."""
    quantum_processor = QuantumProcessor(config=QUANTUM_CONFIG)
    return quantum_processor


if __name__ == "__main__":
    host = os.getenv("API_HOST", "127.0.0.1")  # Default to localhost
    port = int(os.getenv("API_PORT", "8000"))
    uvicorn.run(app, host=host, port=port)



import asyncio
import logging
import time

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from sklearn.datasets import make_classification
from sklearn.metrics import auc, classification_report, roc_curve
from sklearn.model_selection import train_test_split

from ..quantum.hybrid.xgboost_quantum_hybrid import HybridConfig, XGBoostQuantumHybrid

# Setup logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)


async def main():
    """Example usage of XGBoostQuantumHybrid with enhanced capabilities"""
    logger.info("Starting quantum-enhanced XGBoost example...")

    # Generate synthetic data with complex patterns
    logger.info("Generating synthetic data...")
    X, y = make_classification(
        n_samples=2000,  # Increased sample size
        n_features=30,  # More features
        n_informative=25,  # More informative features
        n_redundant=5,
        n_clusters_per_class=2,  # Multiple clusters per class
        random_state=42,
    )

    # Split data
    logger.info("Splitting data into train/test sets...")
    features_train, features_test, labels_train, labels_test = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    # Initialize hybrid model with enhanced configuration
    logger.info("Initializing hybrid model with enhanced configuration...")
    config = HybridConfig(
        # XGBoost parameters
        n_estimators=200,  # More trees
        learning_rate=0.05,  # Lower learning rate for better generalization
        max_depth=5,  # Deeper trees
        objective="binary:logistic",
        eval_metric="auc",
        # Quantum parameters
        n_qubits=6,  # Increased number of qubits
        n_layers=3,  # More layers for deeper quantum processing
        entanglement="full",  # Full entanglement for maximum quantum advantage
        shots=2000,  # More shots for better precision
        optimization_level=3,
        error_correction=True,
        use_advanced_circuits=True,
        quantum_feature_ratio=0.4,  # Increased quantum feature ratio
        # Enhanced features
        use_error_mitigation=True,
        use_quantum_memory=True,
        use_adaptive_entanglement=True,
        quantum_feature_selection=True,
        # Performance parameters
        batch_size=64,  # Larger batch size
        early_stopping_rounds=10,
        use_gpu=True,
    )

    # Create and train model
    logger.info("Creating and training hybrid model...")
    model = XGBoostQuantumHybrid(config=config)

    # Train model and get metrics
    start_time = time.time()
    metrics = await model.train(features_train, labels_train, validation_split=0.2)
    training_time = time.time() - start_time

    logger.info(f"Training completed in {training_time:.2f} seconds")
    logger.info("Training metrics:")
    logger.info(f"- Train AUC: {metrics['train']['auc']:.4f}")
    logger.info(f"- Validation AUC: {metrics['val']['auc']:.4f}")
    logger.info(f"- Quantum metrics: {metrics['quantum_metrics']}")

    # Make predictions
    logger.info("Making predictions...")
    y_pred = await model.predict(features_test)
    y_pred_proba = await model.predict(features_test, return_proba=True)

    # Print classification report
    logger.info("\nClassification Report:")
    print(classification_report(labels_test, y_pred))

    # Get feature importance
    feature_importance = model.get_feature_importance()

    # Plot feature importance
    plt.figure(figsize=(12, 6))
    features = [f"Feature {i}" for i in range(len(feature_importance))]
    importance_df = pd.DataFrame(
        {"Feature": features, "Importance": feature_importance}
    ).sort_values("Importance", ascending=False)

    sns.barplot(data=importance_df, x="Importance", y="Feature")
    plt.title("Feature Importance (Hybrid Model)")
    plt.tight_layout()
    plt.savefig("feature_importance.png")
    logger.info("Feature importance plot saved as 'feature_importance.png'")

    # Plot ROC curve
    plt.figure(figsize=(8, 8))
    fpr, tpr, _ = roc_curve(labels_test, y_pred_proba)
    roc_auc = auc(fpr, tpr)

    plt.plot(
        fpr, tpr, color="darkorange", lw=2, label=f"ROC curve (AUC = {roc_auc:.2f})"
    )
    plt.plot([0, 1], [0, 1], color="navy", lw=2, linestyle="--")
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    plt.title("Receiver Operating Characteristic")
    plt.legend(loc="lower right")
    plt.tight_layout()
    plt.savefig("roc_curve.png")
    logger.info("ROC curve plot saved as 'roc_curve.png'")

    # Plot training history
    training_history = model.get_training_history()
    plt.figure(figsize=(10, 6))

    # Extract metrics
    iterations = [h["iteration"] for h in training_history]
    train_auc = [h["metrics"]["train-auc"] for h in training_history]
    val_auc = [h["metrics"]["val-auc"] for h in training_history]

    plt.plot(iterations, train_auc, label="Train AUC")
    plt.plot(iterations, val_auc, label="Validation AUC")
    plt.xlabel("Iteration")
    plt.ylabel("AUC")
    plt.title("Training History")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("training_history.png")
    logger.info("Training history plot saved as 'training_history.png'")

    # Plot quantum metrics
    quantum_metrics = model.get_quantum_metrics()
    plt.figure(figsize=(10, 6))

    metrics_to_plot = ["error_rate", "entanglement_quality", "memory_usage"]
    for metric in metrics_to_plot:
        if metric in quantum_metrics:
            plt.plot(
                iterations, [quantum_metrics[metric]] * len(iterations), label=metric
            )

    plt.xlabel("Iteration")
    plt.ylabel("Value")
    plt.title("Quantum Metrics")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("quantum_metrics.png")
    logger.info("Quantum metrics plot saved as 'quantum_metrics.png'")


if __name__ == "__main__":
    asyncio.run(main())



from datetime import datetime, timezone
from functools import lru_cache
from typing import List

from pydantic import ConfigDict
from pydantic_settings import BaseSettings


class TestSettings(BaseSettings):
    # Application Settings
    APP_NAME: str = "Bleu.js Test"
    VERSION: str = "0.1.0"
    DEBUG: bool = True

    # Database Settings
    DATABASE_URL: str = "sqlite:///./test.db"
    DB_HOST: str = "localhost"
    DB_PORT: int = 5432
    DB_NAME: str = "test_db"
    DB_USER: str = "test_user"
    DB_PASSWORD: str = "test_password"

    # JWT Settings
    JWT_SECRET_KEY: str = "test_secret_key"
    JWT_ALGORITHM: str = "HS256"
    ACCESS_TOKEN_EXPIRE_MINUTES: int = 30

    # Stripe Settings
    STRIPE_SECRET_KEY: str = "test_stripe_key"
    STRIPE_PUBLISHABLE_KEY: str = "test_stripe_publishable_key"
    STRIPE_WEBHOOK_SECRET: str = "test_webhook_secret"

    # Product IDs
    CORE_PLAN_ID: str = "test_core_plan"
    ENTERPRISE_PLAN_ID: str = "test_enterprise_plan"

    # Rate Limiting
    RATE_LIMIT_CORE: int = 100
    RATE_LIMIT_ENTERPRISE: int = 1000
    RATE_LIMIT_PERIOD: int = 3600  # 1 hour in seconds

    # Security
    CORS_ORIGINS: str = "http://localhost:3000"
    ALLOWED_HOSTS: str = "localhost,127.0.0.1"

    # Testing
    TESTING: bool = True

    # API Settings
    API_VERSION: str = "v1"
    API_PREFIX: str = "/api"

    # Test User Settings
    TEST_USER_EMAIL: str = "test@example.com"
    TEST_USER_PASSWORD: str = "test_password_123"
    TEST_API_KEY: str = "test_api_key_123"

    # Test Database Settings
    TEST_DB_HOST: str = "localhost"
    TEST_DB_PORT: str = "5432"
    TEST_DB_NAME: str = "test_db"
    TEST_DB_USER: str = "test_user"
    TEST_DB_PASSWORD: str = "test_db_password_123"

    # Test API Settings
    TEST_API_HOST: str = "localhost"
    TEST_API_PORT: str = "8000"
    TEST_RATE_LIMIT: str = "100"
    TEST_RATE_LIMIT_WINDOW: str = "3600"

    model_config = ConfigDict(env_file=".env.test", extra="allow")


@lru_cache()
def get_test_settings() -> TestSettings:
    return TestSettings()



import os

from dotenv import load_dotenv

load_dotenv()

# Database configuration
DB_CONFIG = {
    "host": os.getenv("DB_HOST", "your-rds-endpoint"),  # Using RDS endpoint
    "port": os.getenv("DB_PORT", "5432"),
    "database": os.getenv("DB_NAME", "bleu_js"),
    "user": os.getenv("DB_USER", "postgres"),
    "password": os.getenv("DB_PASSWORD", "your_secure_db_password"),
}

# Create database URL
if os.getenv("TESTING") == "true":
    DATABASE_URL = "sqlite:///./test.db"
else:
    DATABASE_URL = f"postgresql://{DB_CONFIG['user']}:{DB_CONFIG['password']}@{DB_CONFIG['host']}:{DB_CONFIG['port']}/{DB_CONFIG['database']}"



"""Rate limiting service module."""

from datetime import datetime, timedelta
from typing import Dict, Optional

from fastapi import HTTPException, status
from redis.asyncio import Redis

from src.config import settings


class RateLimitingService:
    """Service for handling rate limiting."""

    def __init__(self, redis: Redis) -> None:
        """Initialize rate limiting service.

        Args:
            redis: Redis client instance
        """
        self.redis = redis
        self.window = settings.RATE_LIMIT_WINDOW
        self.max_requests = settings.RATE_LIMIT_MAX_REQUESTS

    async def check_rate_limit(self, key: str) -> bool:
        """Check if request is rate limited.

        Args:
            key: Rate limit key (e.g. IP address)

        Returns:
            bool: True if rate limited, False otherwise

        Raises:
            HTTPException: If rate limit is exceeded
        """
        current = await self.redis.get(key)
        if not current:
            await self.redis.setex(key, self.window, 1)
            return False

        count = int(current)
        if count >= self.max_requests:
            retry_after = await self.get_retry_after(key)
            headers = {"Retry-After": str(retry_after)} if retry_after else {}
            raise HTTPException(
                status_code=status.HTTP_429_TOO_MANY_REQUESTS,
                detail="Too many requests",
                headers=headers,
            )

        await self.redis.incr(key)
        return False

    async def get_retry_after(self, key: str) -> Optional[int]:
        """Get retry after time in seconds.

        Args:
            key: Rate limit key

        Returns:
            Optional[int]: Seconds until rate limit resets
        """
        ttl = await self.redis.ttl(key)
        return max(0, ttl) if ttl > 0 else None

    async def get_rate_limit_status(self, key: str) -> Dict[str, int]:
        """Get rate limit status.

        Args:
            key: Rate limit key

        Returns:
            Dict[str, int]: Rate limit status with remaining requests and reset time
        """
        current = await self.redis.get(key)
        ttl = await self.get_retry_after(key) or 0

        remaining = self.max_requests - (int(current) if current else 0)
        reset = int((datetime.now() + timedelta(seconds=ttl)).timestamp())

        return {
            "remaining": remaining,
            "reset": reset,
        }


# Create singleton instance
rate_limiter: Optional[RateLimitingService] = None



import logging
import smtplib
from datetime import datetime, timedelta, timezone
from email.mime.text import MIMEText
from typing import Optional

import httpx
from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer
from jose import JWTError, jwt
from passlib.context import CryptContext
from sqlalchemy.orm import Session

from src.config.settings import settings
from src.database import get_db
from src.models.subscription import Subscription, SubscriptionPlan
from src.models.user import User, UserCreate, UserResponse

logger = logging.getLogger(__name__)

# Security configuration from settings
SECRET_KEY = settings.JWT_SECRET_KEY
ALGORITHM = settings.JWT_ALGORITHM
ACCESS_TOKEN_EXPIRE_MINUTES = settings.JWT_ACCESS_TOKEN_EXPIRE_MINUTES
REFRESH_SECRET_KEY = settings.JWT_SECRET_KEY  # Using same secret key for refresh tokens

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")


class AuthService:
    def __init__(self, db: Session):
        self.pwd_context = pwd_context
        self.oauth2_scheme = oauth2_scheme
        self.db = db

    def verify_password(self, plain_password: str, hashed_password: str) -> bool:
        return self.pwd_context.verify(plain_password, hashed_password)

    def get_password_hash(self, password: str) -> str:
        return self.pwd_context.hash(password)

    def create_access_token(
        self, data: dict, expires_delta: Optional[timedelta] = None
    ) -> str:
        to_encode = data.copy()
        if expires_delta:
            expire = datetime.now(timezone.utc) + expires_delta
        else:
            expire = datetime.now(timezone.utc) + timedelta(minutes=15)
        to_encode.update({"exp": expire})
        encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
        return encoded_jwt

    def create_refresh_token(self, data: dict):
        to_encode = data.copy()
        expire = datetime.now(timezone.utc) + timedelta(days=7)
        to_encode.update({"exp": expire})
        encoded_jwt = jwt.encode(to_encode, REFRESH_SECRET_KEY, algorithm=ALGORITHM)
        return encoded_jwt

    async def verify_refresh_token(self, token: str):
        try:
            payload = jwt.decode(token, REFRESH_SECRET_KEY, algorithms=[ALGORITHM])
            if datetime.now(timezone.utc) > datetime.fromtimestamp(
                payload["exp"], timezone.utc
            ):
                raise HTTPException(status_code=401, detail="Refresh token has expired")
            return payload
        except JWTError:
            raise HTTPException(status_code=401, detail="Invalid refresh token")

    async def get_current_user(self, token: str = None) -> UserResponse:
        """Get the current authenticated user."""
        credentials_exception = HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
            headers={"WWW-Authenticate": "Bearer"},
        )
        try:
            if not token:
                raise credentials_exception
            payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
            email: str = payload.get("sub")
            if email is None:
                raise credentials_exception
        except JWTError:
            raise credentials_exception

        user = self.db.query(User).filter(User.email == email).first()
        if user is None:
            raise credentials_exception

        return UserResponse.model_validate(user)

    async def create_user(self, user: UserCreate, db: Session) -> UserResponse:
        """Create a new user with subscription."""
        # Check if user exists
        db_user = db.query(User).filter(User.email == user.email).first()
        if db_user:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Email already registered",
            )

        # Get the selected plan
        plan = (
            db.query(SubscriptionPlan)
            .filter(SubscriptionPlan.plan_type == user.plan_type)
            .first()
        )
        if not plan:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail="Invalid plan type"
            )

        # Create user
        hashed_password = self.get_password_hash(user.password)
        trial_end_date = datetime.now(timezone.utc) + timedelta(days=plan.trial_days)

        db_user = User(
            email=user.email,
            hashed_password=hashed_password,
            full_name=user.full_name,
            trial_end_date=trial_end_date,
        )
        db.add(db_user)
        db.commit()
        db.refresh(db_user)

        # Create subscription
        subscription = Subscription(
            user_id=db_user.id,
            plan_id=plan.id,
            plan_type=user.plan_type,
            status="active",
            current_period_start=datetime.now(timezone.utc),
            current_period_end=trial_end_date,
        )
        db.add(subscription)
        db.commit()

        # Send verification email
        await self.send_verification_email(user.email)

        return UserResponse.model_validate(db_user)

    async def verify_email(self, token: str, db: Session) -> bool:
        try:
            payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
            email: str = payload.get("sub")
            if email is None:
                return False

            user = db.query(User).filter(User.email == email).first()
            if user:
                user.is_verified = True
                db.commit()
                return True
            return False
        except JWTError:
            return False

    async def authenticate_user(
        self, email: str, password: str, db: Session
    ) -> Optional[User]:
        user = db.query(User).filter(User.email == email).first()
        if not user:
            return None
        if not self.verify_password(password, user.hashed_password):
            return None
        return user

    async def social_auth(self, provider: str, token: str, db: Session) -> UserResponse:
        if provider == "github":
            user_data = await self.get_github_user(token)
        elif provider == "google":
            user_data = await self.get_google_user(token)
        else:
            raise HTTPException(status_code=400, detail="Unsupported provider")

        # Check if user exists
        db_user = db.query(User).filter(User.email == user_data["email"]).first()
        if db_user:
            return UserResponse.from_orm(db_user)

        # Create new user
        db_user = User(
            email=user_data["email"],
            full_name=user_data["name"],
            plan="cor-e",  # Default to COR-E plan
            is_active=True,
            is_verified=True,  # Social auth users are pre-verified
            trial_end_date=datetime.now(timezone.utc) + timedelta(days=30),
        )
        db.add(db_user)
        db.commit()
        db.refresh(db_user)

        return UserResponse.from_orm(db_user)

    async def get_github_user(self, token: str) -> dict:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                "https://api.github.com/user",
                headers={"Authorization": f"Bearer {token}"},
            )
            if response.status_code != 200:
                raise HTTPException(status_code=400, detail="Invalid GitHub token")

            user_data = response.json()
            return {
                "email": user_data["email"],
                "name": user_data["name"] or user_data["login"],
            }

    async def get_google_user(self, token: str) -> dict:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                "https://www.googleapis.com/oauth2/v3/userinfo",
                headers={"Authorization": f"Bearer {token}"},
            )
            if response.status_code != 200:
                raise HTTPException(status_code=400, detail="Invalid Google token")

            user_data = response.json()
            return {"email": user_data["email"], "name": user_data["name"]}

    async def send_verification_email(self, email: str):
        """Send email verification link to user."""
        try:
            # Create verification token
            token = self.create_access_token(
                data={"sub": email}, expires_delta=timedelta(hours=24)
            )

            # Create verification URL
            verification_url = f"{settings.BASE_URL}/verify-email?token={token}"

            # Email content
            subject = "Verify your Bleu.js account"
            body = f"""
            Welcome to Bleu.js!

            Please verify your email address by clicking the link below:
            {verification_url}

            This link will expire in 24 hours.

            If you did not create this account, please ignore this email.
            """

            # Send email using configured SMTP settings
            with smtplib.SMTP(settings.SMTP_HOST, settings.SMTP_PORT) as server:
                server.starttls()
                server.login(settings.SMTP_USER, settings.SMTP_PASSWORD)

                msg = MIMEText(body)
                msg["Subject"] = subject
                msg["From"] = settings.SMTP_USER
                msg["To"] = email

                server.send_message(msg)

            logger.info(f"Verification email sent to {email}")
            return True
        except Exception as e:
            logger.error(f"Failed to send verification email to {email}: {str(e)}")
            return False


# Create a singleton instance
auth_service = AuthService(get_db())



"""Rate limiting service module."""

import time
from typing import Optional

from redis.asyncio import Redis


class RateLimitingService:
    """Service for managing rate limiting."""

    def __init__(self, redis: Redis) -> None:
        """Initialize rate limiting service.

        Args:
            redis: Redis client instance
        """
        self.redis = redis

    async def check_rate_limit(
        self,
        client_id: str,
        max_requests: int,
        window_seconds: int,
    ) -> bool:
        """Check if client has exceeded rate limit.

        Args:
            client_id: Client identifier
            max_requests: Maximum number of requests per window
            window_seconds: Window size in seconds

        Returns:
            bool: True if client has not exceeded rate limit
        """
        # Get current window key
        window_key = self._get_window_key(client_id, window_seconds)

        # Get current request count
        count = await self.redis.get(window_key)
        if not count:
            # First request in window
            pipeline = self.redis.pipeline()
            pipeline.set(window_key, 1)
            pipeline.expire(window_key, window_seconds)
            await pipeline.execute()
            return True

        # Check if limit exceeded
        count = int(count)
        if count >= max_requests:
            return False

        # Increment request count
        await self.redis.incr(window_key)
        return True

    async def get_remaining_requests(
        self,
        client_id: str,
        max_requests: int,
        window_seconds: int,
    ) -> int:
        """Get remaining requests for client.

        Args:
            client_id: Client identifier
            max_requests: Maximum number of requests per window
            window_seconds: Window size in seconds

        Returns:
            int: Number of remaining requests
        """
        # Get current window key
        window_key = self._get_window_key(client_id, window_seconds)

        # Get current request count
        count = await self.redis.get(window_key)
        if not count:
            return max_requests

        # Calculate remaining requests
        count = int(count)
        return max(0, max_requests - count)

    async def get_window_reset_time(
        self,
        client_id: str,
        window_seconds: int,
    ) -> int:
        """Get time until rate limit window resets.

        Args:
            client_id: Client identifier
            window_seconds: Window size in seconds

        Returns:
            int: Seconds until window resets
        """
        # Get current window key
        window_key = self._get_window_key(client_id, window_seconds)

        # Get TTL of window
        ttl = await self.redis.ttl(window_key)
        if ttl < 0:
            return window_seconds

        return ttl

    async def clear_rate_limit(
        self,
        client_id: str,
        window_seconds: int,
    ) -> None:
        """Clear rate limit for client.

        Args:
            client_id: Client identifier
            window_seconds: Window size in seconds
        """
        # Get current window key
        window_key = self._get_window_key(client_id, window_seconds)

        # Delete window key
        await self.redis.delete(window_key)

    def _get_window_key(self, client_id: str, window_seconds: int) -> str:
        """Get Redis key for rate limit window.

        Args:
            client_id: Client identifier
            window_seconds: Window size in seconds

        Returns:
            str: Redis key
        """
        # Calculate window start time
        window_start = int(time.time() / window_seconds) * window_seconds
        return f"rate_limit:{client_id}:{window_start}"



from datetime import datetime, timedelta, timezone

import pytest
from fastapi import HTTPException
from sqlalchemy.orm import Session

from models.rate_limit import RateLimit
from models.subscription import APIToken, APITokenCreate
from models.user import User
from services.api_token_service import APITokenService
from services.rate_limiting_service import RateLimitingService
from services.token_manager import TokenManager


@pytest.fixture
def test_token(db: Session, test_user):
    token = APIToken(
        id="test_token_id",
        user_id=test_user.id,
        name="Test Token",
        token="test_token_value",
        is_active=True,
    )
    db.add(token)
    db.commit()
    return token


@pytest.mark.asyncio
async def test_validate_token_success(db_session, test_api_token):
    """Test successful token validation."""
    token = await APITokenService.validate_token(
        token=test_api_token.token, db=db_session
    )

    assert token is not None
    assert token.id == test_api_token.id
    assert token.is_active is True
    assert token.last_used is not None


@pytest.mark.asyncio
async def test_validate_token_not_found(db_session):
    """Test validation of non-existent token."""
    token = await APITokenService.validate_token(
        token="nonexistent-token", db=db_session
    )

    assert token is None


@pytest.mark.asyncio
async def test_validate_token_expired(db_session, test_user):
    """Test validation of expired token."""
    expired_token = APIToken(
        id="expired-token-id",
        user_id=test_user.id,
        name="Expired Token",
        token="expired-token-value",
        is_active=True,
        created_at=datetime.now(timezone.utc) - timedelta(days=31),
        expires_at=datetime.now(timezone.utc) - timedelta(days=1),
    )
    db_session.add(expired_token)
    db_session.commit()

    token = await APITokenService.validate_token(
        token="expired-token-value", db=db_session
    )

    assert token is None

    # Verify token is deactivated
    expired_token = (
        db_session.query(APIToken).filter(APIToken.id == "expired-token-id").first()
    )
    assert expired_token.is_active is False


@pytest.mark.asyncio
async def test_validate_token_inactive(db_session, test_api_token):
    """Test validation of inactive token."""
    test_api_token.is_active = False
    db_session.commit()

    token = await APITokenService.validate_token(
        token=test_api_token.token, db=db_session
    )

    assert token is None


@pytest.mark.asyncio
async def test_validate_token_rate_limit(db_session, test_api_token):
    """Test token validation with rate limiting."""
    # Set up rate limiting
    test_api_token.rate_limit = 100
    test_api_token.rate_limit_period = 60  # 1 minute
    db_session.commit()

    # Make multiple requests
    for _ in range(100):
        token = await APITokenService.validate_token(
            token=test_api_token.token, db=db_session
        )
        assert token is not None

    # Exceed rate limit
    token = await APITokenService.validate_token(
        token=test_api_token.token, db=db_session
    )
    assert token is None


@pytest.mark.asyncio
async def test_validate_token_with_subscription(
    db_session, test_user, test_subscription
):
    """Test token validation with subscription."""
    token_data = APITokenCreate(
        name="Test Token", expires_at=datetime.now(timezone.utc) + timedelta(days=30)
    )

    token = await APITokenService.create_token(
        user_id=test_user.id, token_data=token_data, db=db_session
    )

    validated_token = await APITokenService.validate_token(
        token=token.token, db=db_session
    )

    assert validated_token is not None
    assert validated_token.id == token.id
    assert validated_token.is_active is True
    assert validated_token.last_used is not None


async def test_validate_token_without_subscription(db: Session, test_user):
    token = APIToken(
        id="no_sub_token_id",
        user_id=test_user.id,
        name="No Sub Token",
        token="no_sub_token_value",
        is_active=True,
    )
    db.add(token)
    db.commit()

    validated_token = APITokenService.validate_token(token.token, db)
    assert validated_token is None


async def test_validate_token_rate_limit_exceeded(
    db: Session, test_user, test_subscription, test_token
):
    # Set rate limit to maximum
    rate_limit = db.query(RateLimit).filter(RateLimit.user_id == test_user.id).first()
    if rate_limit is None:
        raise ValueError("Rate limit not found")
    rate_limit.calls_count = test_subscription.plan.api_calls_limit
    db.commit()

    # Try to validate token
    validated_token = APITokenService.validate_token(test_token.token, db)
    assert validated_token is None


async def test_validate_token_after_rate_limit_reset(
    db: Session, test_user, test_subscription, test_token
):
    # Set rate limit to maximum
    rate_limit = db.query(RateLimit).filter(RateLimit.user_id == test_user.id).first()
    if rate_limit is None:
        raise ValueError("Rate limit not found")
    rate_limit.calls_count = test_subscription.plan.api_calls_limit
    rate_limit.last_reset = datetime.now(timezone.utc) - timedelta(minutes=2)
    db.commit()

    # Try to validate token (should work after reset)
    validated_token = APITokenService.validate_token(test_token.token, db)
    assert validated_token is not None
    assert validated_token.id == test_token.id


def test_token_usage_tracking():
    token = TokenManager.get_token("test_token")
    if token is not None:
        assert token.calls_count == 0
        assert token.last_reset is not None


def test_token_reset():
    token = TokenManager.get_token("test_token")
    if token is not None:
        token.calls_count = 100
        TokenManager.reset_token("test_token")
        assert token.calls_count == 0
        assert token.last_reset is not None



"""Quantum circuit implementation for feature processing."""

import logging
import time
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray
from qiskit import ClassicalRegister
from qiskit import QuantumCircuit as QiskitCircuit
from qiskit import QuantumRegister
from qiskit.circuit.library import EfficientSU2, TwoLocal
from qiskit.primitives import Sampler
from qiskit.quantum_info import Statevector
from qiskit_aer.noise import NoiseModel
from qiskit_algorithms.optimizers import SPSA
from qiskit_machine_learning.algorithms import NeuralNetworkClassifier
from qiskit_machine_learning.neural_networks import SamplerQNN

# Constants
CIRCUIT_NOT_INITIALIZED_ERROR = "Circuit not initialized"

logger = logging.getLogger(__name__)


class CircuitError(Exception):
    """Custom exception for circuit-related errors."""

    pass


class QuantumCircuit:
    """Quantum circuit for feature processing and optimization"""

    def __init__(
        self,
        n_qubits: int = 4,
        n_layers: int = 2,
        entanglement: str = "full",
        use_advanced_circuits: bool = True,
        use_error_mitigation: bool = True,
        use_quantum_memory: bool = True,
        use_adaptive_entanglement: bool = True,
    ) -> None:
        """Initialize quantum circuit with specified parameters.

        Args:
            n_qubits: Number of qubits in the circuit
            n_layers: Number of layers in the circuit
            entanglement: Entanglement pattern ("full", "linear", or "circular")
            use_advanced_circuits: Whether to use advanced circuit components
            use_error_mitigation: Whether to use error mitigation
            use_quantum_memory: Whether to use quantum memory
            use_adaptive_entanglement: Whether to use adaptive entanglement
        """
        self.n_qubits = n_qubits
        self.n_layers = n_layers
        self.entanglement = entanglement
        self.use_advanced_circuits = use_advanced_circuits
        self.use_error_mitigation = use_error_mitigation
        self.use_quantum_memory = use_quantum_memory
        self.use_adaptive_entanglement = use_adaptive_entanglement

        # Initialize quantum registers
        self.qr = QuantumRegister(n_qubits, "q")
        self.cr = ClassicalRegister(n_qubits, "c")

        # Initialize circuit
        self._initialize_circuit()
        self.feature_map: Optional[TwoLocal] = None
        self.ansatz: Optional[EfficientSU2] = None
        self.qnn: Optional[SamplerQNN] = None
        self.classifier: Optional[NeuralNetworkClassifier] = None

        # Initialize quantum memory
        self.quantum_memory: Dict[str, Any] = {}

        # Initialize error mitigation
        self.noise_model: Optional[NoiseModel] = (
            self._get_noise_model() if use_error_mitigation else None
        )

        # Initialize optimizer
        self.optimizer = SPSA(maxiter=100)

        # Initialize metrics
        self.metrics: Dict[str, float] = {
            "circuit_depth": 0.0,
            "circuit_size": 0.0,
            "circuit_width": 0.0,
            "num_parameters": 0.0,
            "optimization_score": 0.0,
            "error_rate": 0.0,
            "entanglement_quality": 0.0,
            "memory_usage": 0.0,
        }

        self.rng = np.random.default_rng(seed=42)  # Fixed seed for reproducibility

        # Build the circuit
        self.build_circuit()

    def _get_noise_model(self) -> NoiseModel:
        """Get noise model for error mitigation"""
        noise_model = NoiseModel()
        # Add depolarizing noise
        noise_model.add_all_qubit_quantum_error(
            noise_model.depolarizing_error(0.01, 1), ["h", "x", "y", "z", "s", "t"]
        )
        # Add readout error
        noise_model.add_all_qubit_readout_error([[0.9, 0.1], [0.1, 0.9]])
        return noise_model

    def _select_quantum_features(self, features: np.ndarray) -> np.ndarray:
        """Select features using quantum feature selection"""
        # Calculate feature importance using quantum state
        importance_scores = np.zeros(features.shape[1])
        for i in range(features.shape[1]):
            # Create quantum state for feature
            state = np.zeros(2**self.n_qubits)
            state[0] = np.mean(features[:, i])
            state[1] = np.std(features[:, i])
            state = state / np.linalg.norm(state)

            # Measure entanglement
            importance_scores[i] = self._measure_entanglement(state)

        # Select top features
        n_selected = min(features.shape[1], self.n_qubits)
        selected_indices = np.argsort(importance_scores)[-n_selected:]
        return features[:, selected_indices]

    def _measure_entanglement(self, state: np.ndarray) -> float:
        """Measure entanglement of quantum state"""
        # Calculate von Neumann entropy
        density_matrix = np.outer(state, state.conj())
        eigenvalues = np.linalg.eigvals(density_matrix)
        eigenvalues = eigenvalues[eigenvalues > 0]
        return -np.sum(eigenvalues * np.log2(eigenvalues))

    def _update_entanglement_pattern(self, features: np.ndarray) -> None:
        """Update entanglement pattern based on feature correlations"""
        if not self.use_adaptive_entanglement or self.circuit is None:
            return

        # Calculate feature correlations
        correlations = np.corrcoef(features.T)

        # Update entanglement pattern
        for i in range(self.n_qubits):
            for j in range(i + 1, self.n_qubits):
                if abs(correlations[i, j]) > 0.5:  # Strong correlation threshold
                    self.circuit.cx(self.qr[i], self.qr[j])

    def build_circuit(self) -> bool:
        """Build the quantum circuit"""
        try:
            # Create base circuit
            self.circuit = QiskitCircuit(self.qr, self.cr)

            if self.use_advanced_circuits:
                # Use advanced circuit components
                self._build_advanced_circuit()
            else:
                # Use basic circuit
                self._build_basic_circuit()

            # Update metrics
            self._update_circuit_metrics()

            return True

        except RuntimeError as e:
            print(f"Error building circuit: {str(e)}")
            return False
        except ValueError as e:
            print(f"Invalid circuit parameters: {str(e)}")
            return False
        except ImportError as e:
            print(f"Failed to import required dependencies: {str(e)}")
            return False

    def _build_advanced_circuit(self) -> None:
        """Build advanced quantum circuit with variational ansatz."""
        if self.circuit is None:
            raise RuntimeError(CIRCUIT_NOT_INITIALIZED_ERROR)

        # Create feature map
        self.feature_map = TwoLocal(
            self.n_qubits,
            "ry",
            "cz",
            entanglement=self.entanglement,
            reps=1,
            parameter_prefix="fm",
        )

        # Create variational ansatz
        self.ansatz = EfficientSU2(
            self.n_qubits,
            entanglement=self.entanglement,
            reps=self.n_layers,
            parameter_prefix="an",
        )

        # Combine circuits
        self.circuit.compose(self.feature_map, inplace=True)
        self.circuit.compose(self.ansatz, inplace=True)

        # Add measurements
        self.circuit.measure(self.qr, self.cr)

    def _build_basic_circuit(self) -> None:
        """Build basic quantum circuit."""
        if self.circuit is None:
            raise RuntimeError(CIRCUIT_NOT_INITIALIZED_ERROR)

        # Add basic gates
        for i in range(self.n_qubits):
            self.circuit.h(i)  # Hadamard gates for superposition
            if i < self.n_qubits - 1:
                self.circuit.cx(i, i + 1)  # CNOT gates for entanglement

        # Add measurements
        self.circuit.measure(self.qr, self.cr)

    def _update_circuit_metrics(self) -> None:
        """Update circuit metrics."""
        if self.circuit is None:
            raise RuntimeError(CIRCUIT_NOT_INITIALIZED_ERROR)

        self.metrics["circuit_depth"] = float(self.circuit.depth())
        self.metrics["circuit_size"] = float(self.circuit.size())
        self.metrics["circuit_width"] = float(self.circuit.width())
        self.metrics["num_parameters"] = float(len(self.circuit.parameters))

    async def process_features(
        self, features: NDArray[np.float64]
    ) -> Optional[NDArray[np.float64]]:
        """Process features using quantum circuit.

        Args:
            features: Input features to process

        Returns:
            Processed features or None if processing fails
        """
        try:
            if self.use_error_mitigation:
                processed = self._process_with_error_mitigation(features)
            else:
                processed = self._process_basic(features)

            if self.use_quantum_memory:
                self._update_quantum_memory(features, processed)

            return processed

        except Exception as e:
            logger.error(f"Error processing features: {str(e)}")
            return None

    def _process_with_error_mitigation(self, features: NDArray[np.float64]) -> float:
        """Process features with error mitigation.

        Args:
            features: Input features to process

        Returns:
            float: Processed result with error mitigation
        """
        try:
            result = self._run_circuit_with_mitigation(features)
            return float(result)  # Explicit cast to float
        except Exception as e:
            logger.error(f"Error in error mitigation processing: {e}")
            raise CircuitError("Failed to process with error mitigation") from e

    def _process_basic(self, features: NDArray[np.float64]) -> NDArray[np.float64]:
        """Process features without error mitigation.

        Args:
            features: Input features to process

        Returns:
            Processed features
        """
        if self.circuit is None:
            raise RuntimeError(CIRCUIT_NOT_INITIALIZED_ERROR)

        # Select quantum features
        quantum_features = self._select_quantum_features(features)

        # Process features
        sampler = Sampler()
        job = sampler.run(self.circuit, quantum_features)
        result = job.result()

        return result.quasi_dists[0]

    def _update_quantum_memory(
        self, features: np.ndarray, processed: np.ndarray
    ) -> None:
        """Update quantum memory with processed features"""
        if self.quantum_memory is None:
            return

        # Store feature patterns
        pattern = np.concatenate([features.flatten(), processed.flatten()])
        self.quantum_memory[hash(str(pattern))] = {
            "features": features,
            "processed": processed,
            "timestamp": time.time(),
        }

    def _calculate_error_rate(self) -> float:
        """Calculate current error rate"""
        if not self.use_error_mitigation or self.noise_model is None:
            return 0.0

        # Use noise model to estimate error rate
        return self.noise_model.depolarizing_error(0.01, 1).probability

    def get_circuit_info(self) -> Dict[str, Any]:
        """Get circuit information and metrics"""
        if self.circuit is None:
            raise RuntimeError(CIRCUIT_NOT_INITIALIZED_ERROR)

        return {
            "n_qubits": self.n_qubits,
            "n_layers": self.n_layers,
            "entanglement": self.entanglement,
            "metrics": self.metrics.copy(),
            "circuit_draw": self.circuit.draw() if self.circuit else None,
            "error_mitigation": self.use_error_mitigation,
            "quantum_memory": bool(self.quantum_memory),
            "adaptive_entanglement": self.use_adaptive_entanglement,
        }

    def apply_random_gate(self, qubit: int) -> None:
        """Apply a random quantum gate to a qubit."""
        if self.circuit is None:
            raise RuntimeError(CIRCUIT_NOT_INITIALIZED_ERROR)

        if qubit < 0 or qubit >= self.n_qubits:
            raise ValueError(f"Invalid qubit index: {qubit}")

        gate_type = self.rng.choice(["H", "X", "Y", "Z", "S", "T"])

        # Apply the selected gate
        if gate_type == "H":
            self.circuit.h(self.qr[qubit])
        elif gate_type == "X":
            self.circuit.x(self.qr[qubit])
        elif gate_type == "Y":
            self.circuit.y(self.qr[qubit])
        elif gate_type == "Z":
            self.circuit.z(self.qr[qubit])
        elif gate_type == "S":
            self.circuit.s(self.qr[qubit])
        elif gate_type == "T":
            self.circuit.t(self.qr[qubit])

    def apply_random_rotation(self, qubit: int) -> None:
        """Apply a random rotation gate to a qubit."""
        if self.circuit is None:
            raise RuntimeError(CIRCUIT_NOT_INITIALIZED_ERROR)

        if qubit < 0 or qubit >= self.n_qubits:
            raise ValueError(f"Invalid qubit index: {qubit}")

        angle = self.rng.uniform(0, 2 * np.pi)
        self.circuit.ry(angle, self.qr[qubit])

    def _initialize_circuit(self) -> None:
        """Initialize the quantum circuit with registers."""
        self.circuit = QiskitCircuit(self.qr, self.cr)

    def _update_metrics(self) -> None:
        self.metrics["circuit_depth"] = self.circuit.depth()
        self.metrics["circuit_size"] = self.circuit.size()
        self.metrics["circuit_width"] = self.circuit.width()
        self.metrics["num_parameters"] = len(self.circuit.parameters)
        self.metrics["error_rate"] = self._calculate_error_rate()

    def add_gate(
        self, gate_name: str, qubits: List[int], params: Optional[List[float]] = None
    ) -> None:
        if params is None:
            params = []

        if gate_name == "h":
            self.circuit.h(qubits[0])
        elif gate_name == "x":
            self.circuit.x(qubits[0])
        elif gate_name == "y":
            self.circuit.y(qubits[0])
        elif gate_name == "z":
            self.circuit.z(qubits[0])
        elif gate_name == "cx":
            self.circuit.cx(qubits[0], qubits[1])
        elif gate_name == "rz":
            self.circuit.rz(params[0], qubits[0])
        else:
            raise ValueError(f"Unsupported gate: {gate_name}")

        self._update_metrics()

    def measure(self, qubit: int) -> int:
        # Create a new classical register for measurement
        cr = self.circuit.add_classical_register(1)
        self.circuit.measure(qubit, cr[0])

        # Execute the circuit and get the measurement result
        from qiskit import Aer, execute

        backend = Aer.get_backend("qasm_simulator")
        job = execute(self.circuit, backend, shots=1)
        result = job.result().get_counts()

        # Extract the measurement outcome
        measured_state = list(result.keys())[0]
        return int(measured_state[-1])

    def get_state(self) -> NDArray[np.complex128]:
        return Statevector.from_instruction(self.circuit).data

    def optimize(self) -> None:
        # Perform circuit optimization
        self._remove_redundant_gates()
        self._merge_adjacent_gates()
        self._update_metrics()

    def _remove_redundant_gates(self) -> None:
        # Identify and remove redundant gates
        current_layer: List[Any] = []
        for instruction in self.circuit.data:
            gate = instruction[0]
            qubits = instruction[1]
            if self._are_gates_cancellable(current_layer, gate, qubits):
                current_layer.pop()
            else:
                current_layer.append(gate)

    def _merge_adjacent_gates(self) -> None:
        # Merge adjacent gates when possible
        current_layer: List[Any] = []
        next_layer: List[Any] = []

        for instruction in self.circuit.data:
            gate = instruction[0]
            qubits = instruction[1]

            if self._can_merge_gates(current_layer, gate):
                self._merge_gates(current_layer, gate, qubits)
            else:
                next_layer = [(gate, qubits)]
                current_layer = next_layer

    def _are_gates_cancellable(
        self, layer: List[Any], gate: Any, qubits: List[int]
    ) -> bool:
        """Check if gates cancel each other based on quantum circuit properties."""
        if not layer:  # Empty layer
            return False

        last_gate = layer[-1]
        # Check for self-cancelling gate pairs (e.g., two consecutive X gates)
        if (
            last_gate.name == gate.name
            and last_gate.name in ["x", "y", "z", "h"]
            and qubits == last_gate.qubits
        ):
            return True
        return False

    def _can_merge_gates(self, layer: List[Any], gate: Any) -> bool:
        """Check if gates can be merged based on quantum circuit optimization rules."""
        if not layer:  # Empty layer
            return False

        last_gate = layer[-1][0]  # Get gate from (gate, qubits) tuple
        # Check for mergeable rotation gates
        if (
            last_gate.name.startswith("r")
            and gate.name.startswith("r")
            and last_gate.name == gate.name
        ):
            return True
        return False

    def _merge_gates(self, layer: List[Any], gate: Any, qubits: List[int]) -> None:
        # Merge compatible gates
        pass  # Placeholder implementation

    def get_metrics(self) -> Dict[str, float]:
        return self.metrics.copy()

    def __str__(self) -> str:
        circuit_str = f"Circuit(num_qubits={self.n_qubits})\n"
        for i, instruction in enumerate(self.circuit.data):
            gate, qubits = instruction
            circuit_str += f"Gate {i}: {gate.name} on qubits {qubits}\n"
        return circuit_str

    def _run_circuit_with_mitigation(self, features: NDArray[np.float64]) -> float:
        # Placeholder for the actual implementation of _run_circuit_with_mitigation
        # This method should return the processed result with error mitigation
        raise NotImplementedError("Method _run_circuit_with_mitigation not implemented")

    def _process_quantum_features(
        self, features: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Process quantum features through the circuit.

        Args:
            features: Input features array

        Returns:
            NDArray[np.float64]: Processed quantum features
        """
        try:
            result = self._quantum_processor.process_features(features)
            return np.array(result, dtype=np.float64)  # Explicit cast to float64 array
        except Exception as e:
            logger.error(f"Error processing quantum features: {e}")
            raise CircuitError("Failed to process quantum features") from e

    def _get_circuit_output(self) -> NDArray[np.float64]:
        """Get the circuit output as a float64 array.

        Returns:
            NDArray[np.float64]: Circuit output array
        """
        try:
            result = self._quantum_processor.get_output()
            return np.array(result, dtype=np.float64)
        except Exception as e:
            logger.error(f"Error getting circuit output: {e}")
            raise CircuitError("Failed to get circuit output") from e

    def _update_metrics(self, key: str, value: float) -> None:
        """Update metrics dictionary with float values.

        Args:
            key: Metric name
            value: Metric value
        """
        if not isinstance(key, str):
            raise TypeError("Metric key must be a string")
        self._metrics[key] = float(value)  # Ensure value is float

    def _get_complex_amplitudes(self) -> NDArray[np.complex128]:
        """Get complex amplitudes from the circuit.

        Returns:
            NDArray[np.complex128]: Complex amplitudes array
        """
        try:
            result = self._quantum_processor.get_amplitudes()
            return np.array(result, dtype=np.complex128)
        except Exception as e:
            logger.error(f"Error getting complex amplitudes: {e}")
            raise CircuitError("Failed to get complex amplitudes") from e



import logging
import os
from contextlib import contextmanager

from dotenv import load_dotenv
from fastapi import Depends
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import QueuePool, StaticPool

from src.models.declarative_base import Base

# Load environment variables
load_dotenv()

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Create engine with appropriate configuration
if os.getenv("TESTING") == "true":
    from tests.test_config import get_test_settings

    settings = get_test_settings()
    if settings.DATABASE_URL.startswith("sqlite"):
        engine = create_engine(
            settings.DATABASE_URL,
            connect_args={"check_same_thread": False},
            poolclass=StaticPool,
        )
    else:
        engine = create_engine(settings.DATABASE_URL)
else:
    from db_config import DATABASE_URL

    engine = create_engine(DATABASE_URL)

# Session factory
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


def get_db():
    """
    FastAPI dependency for database sessions.
    Ensures the session is properly closed after use and handles rollback on errors.
    """
    db = SessionLocal()
    try:
        yield db
    except Exception as e:
        db.rollback()
        logger.error(f"Database error: {str(e)}")
        raise
    finally:
        db.close()


def init_db():
    """Initialize the database, creating all tables."""
    try:
        Base.metadata.create_all(bind=engine)
        logger.info("Database initialized successfully")
    except Exception as e:
        logger.error(f"Failed to initialize database: {str(e)}")
        raise


def check_db_connection():
    """Check if database connection is working."""
    try:
        with engine.connect() as conn:
            conn.execute("SELECT 1")
        logger.info("Database connection successful")
        return True
    except Exception as e:
        logger.error(f"Database connection failed: {str(e)}")
        return False



"""Quantum attention module."""

from typing import List, Optional, Tuple

import numpy as np
import pennylane as qml
from pennylane import numpy as pnp


class Attention:
    """Quantum attention mechanism."""

    def __init__(
        self,
        n_qubits: int,
        n_layers: int = 2,
        learning_rate: float = 0.01,
        shots: Optional[int] = None,
    ) -> None:
        """Initialize quantum attention.

        Args:
            n_qubits: Number of qubits
            n_layers: Number of layers (default: 2)
            learning_rate: Learning rate (default: 0.01)
            shots: Number of shots (default: None)
        """
        self.n_qubits = n_qubits
        self.n_layers = n_layers
        self.learning_rate = learning_rate
        self.shots = shots

        # Initialize device
        self.dev = qml.device("default.qubit", wires=n_qubits, shots=shots)

        # Initialize parameters
        self.params = self._init_parameters()

    def _init_parameters(self) -> np.ndarray:
        """Initialize circuit parameters.

        Returns:
            np.ndarray: Circuit parameters
        """
        # Parameters for each layer
        params_per_layer = self.n_qubits * 3  # Rx, Ry, Rz rotations
        total_params = params_per_layer * self.n_layers

        # Initialize with random values
        return np.random.uniform(low=-np.pi, high=np.pi, size=total_params)

    @qml.qnode(device="default.qubit")
    def _circuit(self, x: np.ndarray, params: np.ndarray) -> List[float]:
        """Quantum circuit for attention mechanism.

        Args:
            x: Input data
            params: Circuit parameters

        Returns:
            List[float]: Measurement results
        """
        # Encode input data
        for i in range(self.n_qubits):
            qml.RY(x[i], wires=i)

        # Apply parameterized layers
        param_idx = 0
        for _ in range(self.n_layers):
            # Single-qubit rotations
            for i in range(self.n_qubits):
                qml.RX(params[param_idx], wires=i)
                qml.RY(params[param_idx + 1], wires=i)
                qml.RZ(params[param_idx + 2], wires=i)
                param_idx += 3

            # Entangling layer
            for i in range(self.n_qubits - 1):
                qml.CNOT(wires=[i, i + 1])
            qml.CNOT(wires=[self.n_qubits - 1, 0])

        # Measure all qubits
        return [qml.expval(qml.PauliZ(i)) for i in range(self.n_qubits)]

    def forward(self, x: np.ndarray) -> np.ndarray:
        """Forward pass through quantum attention.

        Args:
            x: Input data

        Returns:
            np.ndarray: Attention weights
        """
        # Normalize input
        x_norm = x / np.linalg.norm(x)

        # Get quantum measurements
        measurements = self._circuit(x_norm, self.params)

        # Convert to attention weights
        weights = np.array(measurements)
        weights = (weights + 1) / 2  # Map from [-1, 1] to [0, 1]
        weights = weights / np.sum(weights)  # Normalize to sum to 1

        return weights

    def backward(
        self, x: np.ndarray, grad: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Backward pass through quantum attention.

        Args:
            x: Input data
            grad: Gradient from next layer

        Returns:
            Tuple[np.ndarray, np.ndarray]: Input gradients and parameter gradients
        """
        # Normalize input
        x_norm = x / np.linalg.norm(x)

        # Get quantum gradients
        grad_x = qml.grad(self._circuit, argnum=0)(x_norm, self.params)
        grad_params = qml.grad(self._circuit, argnum=1)(x_norm, self.params)

        # Update parameters
        self.params = self.params - self.learning_rate * grad_params

        return grad_x, grad_params

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """Call quantum attention.

        Args:
            x: Input data

        Returns:
            np.ndarray: Attention weights
        """
        return self.forward(x)



import asyncio
import json
import logging
import time
from datetime import datetime, timedelta, timezone
from typing import Dict, List, Optional

import aiohttp
import jwt
from fastapi import Depends, HTTPException
from prometheus_client import Counter, Gauge, Histogram
from sqlalchemy.orm import Session

from ..config import settings
from ..database import get_db
from ..models import APICall, APIUsage, Subscription, User
from ..quantum_py.intelligence.market_intelligence import MarketIntelligence
from ..quantum_py.intelligence.quantum_intelligence import QuantumIntelligence
from ..quantum_py.intelligence.strategic_intelligence import StrategicIntelligence

# Prometheus metrics
api_calls_total = Counter(
    "api_calls_total", "Total number of API calls", ["endpoint", "status"]
)
api_response_time = Histogram(
    "api_response_time_seconds", "API response time in seconds", ["endpoint"]
)
api_errors_total = Counter(
    "api_errors_total", "Total number of API errors", ["endpoint", "error_type"]
)
active_users = Gauge("active_users", "Number of active users")
api_usage_gauge = Gauge("api_usage", "Current API usage", ["user_id", "plan"])


class APIService:
    def __init__(self, db: Session = Depends(get_db)):
        self.db = db
        self.market_intelligence = MarketIntelligence()
        self.strategic_intelligence = StrategicIntelligence()
        self.quantum_intelligence = QuantumIntelligence()
        self.logger = logging.getLogger(__name__)

    async def validate_api_key(self, api_key: str) -> User:
        """Validate API key and return associated user"""
        try:
            payload = jwt.decode(api_key, settings.SECRET_KEY, algorithms=["HS256"])
            user_id = payload.get("sub")
            if not user_id:
                raise HTTPException(status_code=401, detail="Invalid API key")

            user = self.db.query(User).filter(User.id == user_id).first()
            if not user:
                raise HTTPException(status_code=401, detail="User not found")

            if not user.is_active:
                raise HTTPException(status_code=401, detail="User account is inactive")

            return user
        except jwt.InvalidTokenError:
            raise HTTPException(status_code=401, detail="Invalid API key")

    async def check_rate_limit(self, user: User) -> bool:
        """Check if user has exceeded rate limit"""
        current_time = datetime.now(timezone.utc)
        recent_calls = (
            self.db.query(APICall)
            .filter(
                APICall.user_id == user.id,
                APICall.timestamp >= current_time - timedelta(minutes=1),
            )
            .count()
        )

        return recent_calls < user.subscription.rate_limit

    async def check_usage_limit(self, user: User) -> bool:
        """Check if user has exceeded monthly usage limit"""
        current_period_start = user.subscription.current_period_start
        current_period_end = user.subscription.current_period_end

        monthly_usage = (
            self.db.query(APIUsage)
            .filter(
                APIUsage.user_id == user.id,
                APIUsage.timestamp >= current_period_start,
                APIUsage.timestamp <= current_period_end,
            )
            .first()
        )

        if not monthly_usage:
            return True

        return monthly_usage.calls_count < user.subscription.api_calls_limit

    async def track_api_call(
        self, user: User, endpoint: str, response_time: float, status: int
    ):
        """Track API call metrics"""
        # Update Prometheus metrics
        api_calls_total.labels(endpoint=endpoint, status=str(status)).inc()
        api_response_time.labels(endpoint=endpoint).observe(response_time)
        if status >= 400:
            api_errors_total.labels(endpoint=endpoint, error_type=str(status)).inc()

        # Update database
        api_call = APICall(
            user_id=user.id,
            endpoint=endpoint,
            response_time=response_time,
            status=status,
            timestamp=datetime.now(timezone.utc),
        )
        self.db.add(api_call)

        # Update monthly usage
        monthly_usage = (
            self.db.query(APIUsage)
            .filter(
                APIUsage.user_id == user.id,
                APIUsage.timestamp >= user.subscription.current_period_start,
                APIUsage.timestamp <= user.subscription.current_period_end,
            )
            .first()
        )

        if not monthly_usage:
            monthly_usage = APIUsage(
                user_id=user.id, calls_count=0, timestamp=datetime.now(timezone.utc)
            )
            self.db.add(monthly_usage)

        monthly_usage.calls_count += 1
        self.db.commit()

        # Update Prometheus gauge
        api_usage_gauge.labels(
            user_id=str(user.id), plan=user.subscription.plan_type
        ).set(monthly_usage.calls_count)

    async def get_usage_analytics(self, user: User) -> Dict:
        """Get detailed usage analytics for user"""
        current_period_start = user.subscription.current_period_start
        current_period_end = user.subscription.current_period_end

        # Get daily usage
        daily_usage = (
            self.db.query(APICall)
            .filter(
                APICall.user_id == user.id,
                APICall.timestamp >= current_period_start,
                APICall.timestamp <= current_period_end,
            )
            .all()
        )

        # Calculate response time distribution
        response_times = [call.response_time for call in daily_usage]
        response_time_distribution = {
            "<100ms": len([t for t in response_times if t < 0.1]),
            "100-200ms": len([t for t in response_times if 0.1 <= t < 0.2]),
            "200-300ms": len([t for t in response_times if 0.2 <= t < 0.3]),
            "300-400ms": len([t for t in response_times if 0.3 <= t < 0.4]),
            ">400ms": len([t for t in response_times if t >= 0.4]),
        }

        # Get endpoint usage
        endpoint_usage = {}
        for call in daily_usage:
            endpoint_usage[call.endpoint] = endpoint_usage.get(call.endpoint, 0) + 1

        # Get error rates
        error_calls = [call for call in daily_usage if call.status >= 400]
        error_rate = len(error_calls) / len(daily_usage) if daily_usage else 0

        return {
            "total_calls": len(daily_usage),
            "response_time_distribution": response_time_distribution,
            "endpoint_usage": endpoint_usage,
            "error_rate": error_rate,
            "average_response_time": (
                sum(response_times) / len(response_times) if response_times else 0
            ),
        }

    async def get_advanced_analytics(self, user: User) -> Dict:
        """Get advanced analytics for enterprise users"""
        if user.subscription.plan_type != "ENTERPRISE":
            raise HTTPException(
                status_code=403,
                detail="Advanced analytics only available for Enterprise plan",
            )

        # Get market intelligence
        market_data = await self.market_intelligence.get_market_analysis()

        # Get strategic insights
        strategic_insights = await self.strategic_intelligence.get_strategic_insights()

        # Get quantum analysis
        quantum_analysis = await self.quantum_intelligence.get_quantum_analysis()

        return {
            "market_intelligence": market_data,
            "strategic_insights": strategic_insights,
            "quantum_analysis": quantum_analysis,
        }

    async def get_user_dashboard_data(self, user: User) -> Dict:
        """Get all dashboard data for user"""
        usage_analytics = await self.get_usage_analytics(user)

        dashboard_data = {
            "user": {
                "email": user.email,
                "plan_type": user.subscription.plan_type,
                "api_key": user.api_key,
            },
            "subscription": {
                "current_period_start": user.subscription.current_period_start,
                "current_period_end": user.subscription.current_period_end,
                "api_calls_limit": user.subscription.api_calls_limit,
                "rate_limit": user.subscription.rate_limit,
            },
            "usage": {
                "calls_remaining": user.subscription.api_calls_limit
                - usage_analytics["total_calls"],
                "current_usage": usage_analytics["total_calls"],
                "response_time_distribution": usage_analytics[
                    "response_time_distribution"
                ],
                "endpoint_usage": usage_analytics["endpoint_usage"],
                "error_rate": usage_analytics["error_rate"],
                "average_response_time": usage_analytics["average_response_time"],
            },
        }

        # Add advanced analytics for enterprise users
        if user.subscription.plan_type == "ENTERPRISE":
            advanced_analytics = await self.get_advanced_analytics(user)
            dashboard_data["advanced_analytics"] = advanced_analytics

        return dashboard_data

    async def monitor_api_health(self) -> Dict:
        """Monitor overall API health"""
        # Get system metrics
        system_metrics = {
            "uptime": time.time() - self.start_time,
            "active_users": active_users._value.get(),
            "total_api_calls": api_calls_total._value.get(),
            "error_rate": (
                api_errors_total._value.get() / api_calls_total._value.get()
                if api_calls_total._value.get() > 0
                else 0
            ),
        }

        # Get database health
        try:
            self.db.execute("SELECT 1")
            db_health = "healthy"
        except Exception as e:
            db_health = "unhealthy"
            self.logger.error(f"Database health check failed: {str(e)}")

        # Get external service health
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(
                    settings.QUANTUM_SERVICE_URL + "/health"
                ) as response:
                    quantum_health = (
                        "healthy" if response.status == 200 else "unhealthy"
                    )
            except Exception as e:
                quantum_health = "unhealthy"
                self.logger.error(f"Quantum service health check failed: {str(e)}")

        return {
            "system_metrics": system_metrics,
            "database_health": db_health,
            "quantum_service_health": quantum_health,
            "timestamp": datetime.now(timezone.utc).isoformat(),
        }

    async def optimize_api_performance(self) -> Dict:
        """Optimize API performance based on usage patterns"""
        # Get recent API calls
        recent_calls = (
            self.db.query(APICall)
            .filter(
                APICall.timestamp >= datetime.now(timezone.utc) - timedelta(hours=24)
            )
            .all()
        )

        # Analyze response times
        response_times = [call.response_time for call in recent_calls]
        avg_response_time = (
            sum(response_times) / len(response_times) if response_times else 0
        )

        # Identify slow endpoints
        endpoint_times = {}
        for call in recent_calls:
            if call.endpoint not in endpoint_times:
                endpoint_times[call.endpoint] = []
            endpoint_times[call.endpoint].append(call.response_time)

        slow_endpoints = {
            endpoint: sum(times) / len(times)
            for endpoint, times in endpoint_times.items()
            if sum(times) / len(times) > avg_response_time * 1.5
        }

        # Generate optimization recommendations
        recommendations = []
        for endpoint, avg_time in slow_endpoints.items():
            if avg_time > 1.0:  # More than 1 second
                recommendations.append(
                    {
                        "endpoint": endpoint,
                        "issue": "High response time",
                        "recommendation": "Implement caching and optimize database queries",
                    }
                )
            elif avg_time > 0.5:  # More than 500ms
                recommendations.append(
                    {
                        "endpoint": endpoint,
                        "issue": "Moderate response time",
                        "recommendation": "Consider implementing request batching",
                    }
                )

        return {
            "average_response_time": avg_response_time,
            "slow_endpoints": slow_endpoints,
            "recommendations": recommendations,
            "timestamp": datetime.now(timezone.utc).isoformat(),
        }



import enum
import uuid
from datetime import datetime, timedelta, timezone
from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict
from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    String,
)
from sqlalchemy.orm import relationship

from src.models.declarative_base import Base


class PlanType(str, enum.Enum):
    FREE = "FREE"
    CORE = "CORE"
    ENTERPRISE = "ENTERPRISE"


class SubscriptionStatus(str, enum.Enum):
    ACTIVE = "active"
    CANCELLED = "cancelled"
    EXPIRED = "expired"
    TRIAL = "trial"


class Subscription(Base):
    """Database model for user subscriptions."""

    __tablename__ = "subscriptions"
    __table_args__ = {"extend_existing": True}

    id = Column(String, primary_key=True, default=lambda: str(uuid.uuid4()))
    user_id = Column(String, ForeignKey("users.id"), nullable=False)
    plan_id = Column(String, ForeignKey("subscription_plans.id"), nullable=False)
    stripe_subscription_id = Column(String, nullable=True)
    status = Column(
        Enum(SubscriptionStatus), nullable=False, default=SubscriptionStatus.TRIAL
    )
    current_period_start = Column(DateTime(timezone=True), nullable=False)
    current_period_end = Column(DateTime(timezone=True), nullable=False)
    api_calls_remaining = Column(Integer, nullable=False)
    created_at = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
    )
    updated_at = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
    )

    # Relationships
    user = relationship("User", back_populates="subscription")
    plan = relationship("SubscriptionPlan", back_populates="subscriptions")
    api_tokens = relationship(
        "APIToken", back_populates="subscription", cascade="all, delete-orphan"
    )
    customer = relationship("Customer", back_populates="subscription", uselist=False)


class APIToken(Base):
    """Database model for storing API tokens."""

    __tablename__ = "api_tokens"
    __table_args__ = {"extend_existing": True}

    id = Column(String, primary_key=True, default=lambda: str(uuid.uuid4()))
    user_id = Column(String, ForeignKey("users.id"), nullable=False)
    subscription_id = Column(String, ForeignKey("subscriptions.id"), nullable=False)
    customer_id = Column(String, ForeignKey("customers.id"), nullable=True)
    name = Column(String, nullable=False)
    token = Column(String, unique=True, nullable=False)
    is_active = Column(Boolean, default=True)
    created_at = Column(DateTime, default=lambda: datetime.now(timezone.utc))
    last_used_at = Column(DateTime, nullable=True)
    expires_at = Column(DateTime, nullable=True)

    # Relationships
    user = relationship("User", back_populates="api_tokens")
    subscription = relationship("Subscription", back_populates="api_tokens")
    customer = relationship("Customer", back_populates="api_tokens")


class SubscriptionPlan(Base):
    """Database model for subscription plans."""

    __tablename__ = "subscription_plans"
    __table_args__ = {"extend_existing": True}

    id = Column(String, primary_key=True, default=lambda: str(uuid.uuid4()))
    name = Column(String, nullable=False)
    plan_type = Column(Enum(PlanType), nullable=False)
    price = Column(Integer, nullable=False)  # Price in cents
    api_calls_limit = Column(Integer, nullable=False)
    rate_limit = Column(Integer, nullable=False)  # Rate limit per second
    uptime_sla = Column(String, nullable=False)
    support_level = Column(String, nullable=False)
    features = Column(JSON, nullable=False)
    trial_days = Column(Integer, nullable=False, default=30)
    created_at = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
    )
    updated_at = Column(
        DateTime(timezone=True),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
    )

    # Relationships
    subscriptions = relationship("Subscription", back_populates="plan")

    @staticmethod
    def get_core_plan():
        return {
            "name": "COR-E Plan",
            "plan_type": PlanType.CORE,
            "price": 2900,  # $29/month
            "api_calls_limit": 100,
            "rate_limit": 10,  # 10 requests per second
            "uptime_sla": "99.9%",
            "support_level": "standard",
            "features": {
                "core_ai_model_access": True,
                "basic_analytics": True,
                "email_support": True,
                "api_documentation": True,
                "standard_response_time": True,
            },
            "trial_days": 14,
        }

    @staticmethod
    def get_enterprise_plan():
        return {
            "name": "Enterprise Plan",
            "plan_type": PlanType.ENTERPRISE,
            "price": 49900,  # $499/month
            "api_calls_limit": 5000,
            "rate_limit": 100,  # 100 requests per second
            "uptime_sla": "99.99%",
            "support_level": "premium",
            "features": {
                "core_ai_model_access": True,
                "advanced_analytics": True,
                "priority_support": True,
                "dedicated_account_manager": True,
                "custom_model_training": True,
                "custom_integrations": True,
                "sla_guarantees": True,
                "advanced_documentation": True,
            },
            "trial_days": 30,
        }


# Pydantic models for API
class APITokenBase(BaseModel):
    name: str
    expires_at: Optional[datetime] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class APITokenCreate(APITokenBase):
    pass


class APITokenResponse(APITokenBase):
    id: str
    user_id: str
    token: str
    is_active: bool
    last_used_at: Optional[datetime]
    created_at: datetime

    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True)


class SubscriptionPlanBase(BaseModel):
    name: str
    plan_type: PlanType
    price: int
    api_calls_limit: int
    trial_days: int
    features: Dict
    rate_limit: int
    uptime_sla: str
    support_level: str

    model_config = ConfigDict(arbitrary_types_allowed=True)


class SubscriptionPlanCreate(SubscriptionPlanBase):
    pass


class SubscriptionPlanResponse(SubscriptionPlanBase):
    id: str
    created_at: datetime
    updated_at: datetime

    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True)


class SubscriptionBase(BaseModel):
    user_id: str
    plan_id: str
    stripe_subscription_id: Optional[str] = None
    status: str
    current_period_start: datetime
    current_period_end: datetime
    trial_end_date: Optional[datetime]
    api_calls_remaining: int

    model_config = ConfigDict(arbitrary_types_allowed=True)


class SubscriptionCreate(SubscriptionBase):
    pass


class SubscriptionResponse(SubscriptionBase):
    id: str
    created_at: datetime
    updated_at: datetime
    plan: SubscriptionPlanResponse

    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True)



"""Advanced XGBoost training with quantum enhancements."""

import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import numpy as np
import xgboost as xgb

logger = logging.getLogger(__name__)


@dataclass
class TrainingConfig:
    """Configuration for XGBoost training"""

    n_estimators: int = 1000
    learning_rate: float = 0.01
    max_depth: int = 6
    min_child_weight: int = 1
    subsample: float = 0.8
    colsample_bytree: float = 0.8
    objective: str = "binary:logistic"
    tree_method: str = "hist"
    use_gpu: bool = True
    early_stopping_rounds: int = 50


@dataclass
class QuantumConfig:
    """Configuration for quantum enhancements"""

    n_qubits: int = 4
    n_layers: int = 2
    entanglement: str = "linear"
    shots: int = 1000
    optimization_level: int = 3
    error_correction: bool = True


@dataclass
class SecurityConfig:
    """Configuration for security features"""

    encryption: bool = False
    secure_training: bool = False
    differential_privacy: bool = False
    privacy_budget: float = 1.0


class AdvancedModelTrainer:
    """Advanced XGBoost trainer with quantum enhancements and security features."""

    def __init__(
        self,
        training_config: Optional[Dict] = None,
        quantum_config: Optional[Dict] = None,
        security_config: Optional[Dict] = None,
    ):
        # Convert dict configs to dataclass instances
        self.training_config = TrainingConfig(**(training_config or {}))
        self.quantum_config = QuantumConfig(**(quantum_config or {}))
        self.security_config = SecurityConfig(**(security_config or {}))

        # Initialize XGBoost model
        self.model = None
        self.feature_importance = None

        # Initialize logging
        self._setup_logging()

    def _setup_logging(self):
        """Setup logging configuration"""
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        )
        self.logger = logging.getLogger(__name__)

    async def train(self, features: np.ndarray, labels: np.ndarray) -> Dict[str, float]:
        """Train the model with quantum enhancements and security measures."""
        try:
            # Apply security measures
            features_secure, labels_secure = self._apply_security_measures(
                features, labels
            )

            # Apply quantum enhancements
            features_enhanced = await self._apply_quantum_enhancements(features_secure)

            # Train model
            self.model = xgb.XGBClassifier(
                n_estimators=self.training_config.n_estimators,
                learning_rate=self.training_config.learning_rate,
                max_depth=self.training_config.max_depth,
                min_child_weight=self.training_config.min_child_weight,
                subsample=self.training_config.subsample,
                colsample_bytree=self.training_config.colsample_bytree,
                objective=self.training_config.objective,
                tree_method=self.training_config.tree_method,
                use_gpu=self.training_config.use_gpu,
            )

            # Train with early stopping
            self.model.fit(
                features_enhanced,
                labels_secure,
                eval_set=[(features_enhanced, labels_secure)],
                early_stopping_rounds=self.training_config.early_stopping_rounds,
                verbose=False,
            )

            # Calculate metrics
            metrics = self._calculate_metrics(features_enhanced, labels_secure)

            # Store feature importance
            self.feature_importance = self.model.feature_importances_

            return metrics

        except Exception as e:
            self.logger.error(f"Training failed: {str(e)}")
            raise

    async def predict(self, features: np.ndarray) -> np.ndarray:
        """Make predictions using the trained model."""
        try:
            if self.model is None:
                raise ValueError("Model has not been trained yet")

            # Apply quantum enhancements
            features_enhanced = await self._apply_quantum_enhancements(features)

            # Make predictions
            predictions = self.model.predict(features_enhanced)

            return predictions

        except Exception as e:
            self.logger.error(f"Prediction failed: {str(e)}")
            raise

    async def _apply_quantum_enhancements(self, features: np.ndarray) -> np.ndarray:
        """Apply quantum enhancements to features"""
        # Placeholder for quantum feature processing
        # This would be implemented with actual quantum circuits
        return features

    def _apply_security_measures(
        self, features: np.ndarray, labels: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Apply security measures to data"""
        if self.security_config.differential_privacy:
            # Add noise for differential privacy
            rng = np.random.default_rng(seed=42)  # Fixed seed for reproducibility
            noise = rng.normal(
                0, 1 / self.security_config.privacy_budget, features.shape
            )
            features = features + noise

        return features, labels

    def _calculate_metrics(
        self,
        features: np.ndarray,
        labels: np.ndarray,
        eval_set: Optional[List[Tuple[np.ndarray, np.ndarray]]] = None,
    ) -> Dict:
        """Calculate training metrics"""
        metrics = {}

        # Training metrics
        if not self.model:
            raise ValueError("Model not initialized")
        predictions = self.model.predict(features)
        metrics["accuracy"] = np.mean(predictions == labels)

        # Validation metrics if eval_set is provided
        if eval_set:
            if not hasattr(self.model, "evals_result"):
                raise ValueError("Model does not support evaluation results")
            val_metrics = self.model.evals_result()
            metrics.update(
                {
                    "val_logloss": val_metrics["validation_0"]["logloss"][-1],
                    "val_auc": val_metrics["validation_0"]["auc"][-1],
                }
            )

        return metrics



"""Services package."""

from typing import Dict

from fastapi import FastAPI
from redis.asyncio import Redis
from sqlalchemy.orm import Session

from src.services.api_service import APIService
from src.services.api_token_service import APITokenService
from src.services.auth_service import AuthService
from src.services.email_service import EmailService
from src.services.monitoring_service import MonitoringService
from src.services.rate_limiting_service import RateLimitingService
from src.services.stripe_service import StripeService
from src.services.subscription_service import SubscriptionService
from src.services.user_service import UserService


def init_services(app: FastAPI, db: Session, redis: Redis) -> Dict[str, object]:
    """Initialize application services.

    Args:
        app: FastAPI application
        db: Database session
        redis: Redis client

    Returns:
        Dict[str, object]: Dictionary of initialized services
    """
    # Initialize services
    rate_limiting_service = RateLimitingService(redis)
    token_service = APITokenService(db)
    user_service = UserService(db)
    auth_service = AuthService(db)
    api_service = APIService(db)
    monitoring_service = MonitoringService()
    stripe_service = StripeService()
    subscription_service = SubscriptionService(db, stripe_service)
    email_service = EmailService()

    # Store services in app state
    app.state.rate_limiting_service = rate_limiting_service
    app.state.token_service = token_service
    app.state.user_service = user_service
    app.state.auth_service = auth_service
    app.state.api_service = api_service
    app.state.monitoring_service = monitoring_service
    app.state.stripe_service = stripe_service
    app.state.subscription_service = subscription_service
    app.state.email_service = email_service

    # Return services dictionary
    return {
        "rate_limiting_service": rate_limiting_service,
        "token_service": token_service,
        "user_service": user_service,
        "auth_service": auth_service,
        "api_service": api_service,
        "monitoring_service": monitoring_service,
        "stripe_service": stripe_service,
        "subscription_service": subscription_service,
        "email_service": email_service,
    }


__all__ = [
    "APIService",
    "APITokenService",
    "AuthService",
    "EmailService",
    "MonitoringService",
    "RateLimitingService",
    "StripeService",
    "SubscriptionService",
    "UserService",
    "init_services",
]



import os
from datetime import datetime, timedelta, timezone
from functools import lru_cache
from typing import List, Optional

import pytest
from dotenv import load_dotenv
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import StaticPool

from src.config import get_settings
from src.models.customer import RateLimitToken
from src.models.declarative_base import Base
from src.models.rate_limit import RateLimit
from src.models.subscription import APIToken, PlanType, Subscription, SubscriptionPlan
from src.models.user import User

# Load environment variables
load_dotenv(".env.test")

# EC2 Configuration
EC2_HOST = "44.245.223.189"
EC2_USER = "ec2-user"
EC2_KEY_PATH = "bleu-js-key.pem"

# API Gateway Configuration
API_GATEWAY_URL = os.getenv("AWS_API_CONFIG_BASE_URL")
TEST_API_KEY = os.getenv("TEST_API_KEY")
ENTERPRISE_API_KEY = os.getenv("ENTERPRISE_TEST_API_KEY")
AWS_REGION = os.getenv("AWS_REGION")

# SSH Commands
SSH_BASE = f"ssh -i {EC2_KEY_PATH} {EC2_USER}@{EC2_HOST}"
SCP_BASE = f"scp -i {EC2_KEY_PATH}"

# Service check commands
SERVICE_COMMANDS = {
    "cloudwatch": "sudo systemctl status amazon-cloudwatch-agent",
    "nginx": "sudo systemctl status nginx",
}

# Log paths
LOG_PATHS = {
    "cloudwatch": "/var/log/amazon/amazon-cloudwatch-agent/amazon-cloudwatch-agent.log",
    "cloud_init": "/var/log/cloud-init-output.log",
}

# Test configuration constants
TEST_EMAIL = "test@example.com"
TEST_NOREPLY_EMAIL = "noreply@example.com"

# Test database URL
TEST_DATABASE_URL = get_settings().DATABASE_URL


class TestSettings(BaseSettings):
    # Application Settings
    APP_NAME: str = Field(default="Bleu.js Test")
    VERSION: str = Field(default="1.1.4")
    DEBUG: bool = Field(default=True)

    # Database Settings
    DATABASE_URL: str = Field(default="sqlite:///./test.db")

    # JWT Settings
    JWT_SECRET_KEY: str = Field(default="test-secret-key")
    JWT_ALGORITHM: str = Field(default="HS256")
    ACCESS_TOKEN_EXPIRE_MINUTES: int = Field(default=30)

    # Stripe Settings
    STRIPE_SECRET_KEY: str = Field(default="test_stripe_secret_key")
    STRIPE_PUBLISHABLE_KEY: str = Field(default="test_stripe_publishable_key")
    STRIPE_WEBHOOK_SECRET: str = Field(default="test_stripe_webhook_secret")

    # Product IDs
    CORE_PLAN_ID: str = Field(default="test_core_plan_id")
    ENTERPRISE_PLAN_ID: str = Field(default="test_enterprise_plan_id")

    # Rate Limiting
    RATE_LIMIT_CORE: int = Field(default=100)
    RATE_LIMIT_ENTERPRISE: int = Field(default=5000)

    # Security
    CORS_ORIGINS: str = Field(default="http://localhost:3000")
    ALLOWED_HOSTS: str = Field(default="*")

    # Email Settings
    SMTP_HOST: str = Field(default="smtp.test.com")
    SMTP_PORT: int = Field(default=587)
    SMTP_USER: str = Field(default=TEST_EMAIL)
    SMTP_PASSWORD: str = Field(default="test_password")
    FROM_EMAIL: str = Field(default=TEST_NOREPLY_EMAIL)

    model_config = SettingsConfigDict(
        env_file=".env.test",
        case_sensitive=True,
        arbitrary_types_allowed=True,
        extra="allow",
    )

    @property
    def cors_origins_list(self) -> List[str]:
        """Get CORS origins as a list."""
        return [origin.strip() for origin in self.CORS_ORIGINS.split(",")]

    @property
    def allowed_hosts_list(self) -> List[str]:
        """Get allowed hosts as a list."""
        return [host.strip() for host in self.ALLOWED_HOSTS.split(",")]


@lru_cache()
def get_test_settings() -> TestSettings:
    return TestSettings()


# Use the test settings instance
test_settings = get_test_settings()


@pytest.fixture(scope="session")
def engine():
    """Create a test database engine."""
    engine = create_engine(
        "sqlite:///:memory:",
        connect_args={"timeout": 30, "check_same_thread": False},
        poolclass=StaticPool,
        isolation_level="SERIALIZABLE",
    )
    Base.metadata.create_all(engine)
    return engine


@pytest.fixture(scope="function")
def db_session(engine):
    """Create a test database session."""
    connection = engine.connect()
    transaction = connection.begin()
    session = sessionmaker(bind=connection)()

    yield session

    session.close()
    transaction.rollback()
    connection.close()


@pytest.fixture(scope="function")
def test_user(db_session):
    user = User(
        id="test-user-id",
        email="test@example.com",
        hashed_password="test-password",
        is_active=True,
        is_superuser=False,
    )
    db_session.add(user)
    db_session.commit()
    return user


@pytest.fixture(scope="function")
def test_subscription_plan(db_session):
    plan = SubscriptionPlan(
        id="test-plan-id",
        name="Test Plan",
        plan_type=PlanType.CORE,
        price=1000,
        api_calls_limit=100,
        trial_days=30,
        features={"core_ai_model_access": True},
        rate_limit=100,
        uptime_sla="99.9%",
        support_level="standard",
    )
    db_session.add(plan)
    db_session.commit()
    return plan


@pytest.fixture(scope="function")
def test_subscription(db_session, test_user, test_subscription_plan):
    subscription = Subscription(
        id="test-subscription-id",
        user_id=test_user.id,
        plan_id=test_subscription_plan.id,
        plan_type=PlanType.CORE,
        status="active",
        current_period_start=datetime.now(timezone.utc),
        current_period_end=datetime.now(timezone.utc) + timedelta(days=30),
    )
    db_session.add(subscription)
    db_session.commit()
    return subscription


@pytest.fixture(scope="function")
def test_api_token(db_session, test_user, test_subscription):
    token = APIToken(
        id="test-token-id",
        user_id=test_user.id,
        subscription_id=test_subscription.id,
        name="Test Token",
        token="test-token-value",
        is_active=True,
        created_at=datetime.now(timezone.utc),
    )
    db_session.add(token)
    db_session.commit()
    return token


@pytest.fixture(scope="function")
def test_rate_limit(db_session, test_user):
    rate_limit = RateLimit(
        id="test-rate-limit-id",
        user_id=test_user.id,
        endpoint="test_endpoint",
        limit=100,
        period=3600,
        calls_count=0,
        last_reset=datetime.now(timezone.utc),
        current_period_start=datetime.now(timezone.utc),
        last_used=datetime.now(timezone.utc),
    )
    db_session.add(rate_limit)
    db_session.commit()
    return rate_limit



"""User model module."""

from datetime import datetime
from typing import Optional

from pydantic import BaseModel, ConfigDict, EmailStr, Field
from sqlalchemy import Boolean, Column, DateTime, Integer, String
from sqlalchemy.orm import relationship

from src.models.declarative_base import Base
from src.models.subscription import PlanType


class User(Base):
    """User model."""

    __tablename__ = "users"
    __table_args__ = {"extend_existing": True}

    id = Column(Integer, primary_key=True, index=True)
    email = Column(String, unique=True, index=True, nullable=False)
    username = Column(String, unique=True, index=True, nullable=False)
    hashed_password = Column(String, nullable=False)
    full_name = Column(String)
    is_active = Column(Boolean, default=True)
    is_superuser = Column(Boolean, default=False)
    created_at = Column(DateTime, default=datetime.utcnow)
    updated_at = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)
    last_login = Column(DateTime)

    # Relationships
    api_tokens = relationship(
        "APIToken", back_populates="user", cascade="all, delete-orphan"
    )
    subscription = relationship(
        "Subscription", back_populates="user", cascade="all, delete-orphan"
    )
    api_calls = relationship(
        "APICall", back_populates="user", cascade="all, delete-orphan"
    )
    api_usage = relationship(
        "APIUsage", back_populates="user", cascade="all, delete-orphan"
    )
    rate_limits = relationship(
        "RateLimit", back_populates="user", cascade="all, delete-orphan"
    )

    def __repr__(self) -> str:
        """Get string representation of user.

        Returns:
            str: String representation
        """
        return f"<User {self.username}>"

    def to_dict(self) -> dict:
        """Convert user to dictionary.

        Returns:
            dict: User data
        """
        return {
            "id": self.id,
            "email": self.email,
            "username": self.username,
            "full_name": self.full_name,
            "is_active": self.is_active,
            "is_superuser": self.is_superuser,
            "created_at": self.created_at.isoformat() if self.created_at else None,
            "updated_at": self.updated_at.isoformat() if self.updated_at else None,
            "last_login": self.last_login.isoformat() if self.last_login else None,
        }

    @property
    def is_authenticated(self) -> bool:
        """Check if user is authenticated.

        Returns:
            bool: True if user is authenticated
        """
        return True

    @property
    def is_anonymous(self) -> bool:
        """Check if user is anonymous.

        Returns:
            bool: True if user is anonymous
        """
        return False

    def get_id(self) -> str:
        """Get user ID.

        Returns:
            str: User ID
        """
        return str(self.id)

    def update_last_login(self) -> None:
        """Update last login timestamp."""
        self.last_login = datetime.utcnow()

    def check_password(self, password: str) -> bool:
        """Check if password is correct.

        Args:
            password: Password to check

        Returns:
            bool: True if password is correct
        """
        from src.security.password import verify_password

        return verify_password(password, self.hashed_password)

    def set_password(self, password: str) -> None:
        """Set user password.

        Args:
            password: Password to set
        """
        from src.security.password import get_password_hash

        self.hashed_password = get_password_hash(password)

    def has_permission(self, permission: str) -> bool:
        """Check if user has permission.

        Args:
            permission: Permission to check

        Returns:
            bool: True if user has permission
        """
        # Superusers have all permissions
        if self.is_superuser:
            return True

        # TODO: Implement permission checking
        return False

    def has_subscription(self, subscription_type: str) -> bool:
        """Check if user has subscription.

        Args:
            subscription_type: Subscription type to check

        Returns:
            bool: True if user has subscription
        """
        return any(
            subscription.type == subscription_type and subscription.is_active
            for subscription in self.subscriptions
        )

    def get_active_api_token(self) -> Optional[str]:
        """Get active API token.

        Returns:
            Optional[str]: Active API token
        """
        active_token = next(
            (token for token in self.api_tokens if token.is_active),
            None,
        )
        return active_token.token if active_token else None


class UserBase(BaseModel):
    """Base user model."""

    email: EmailStr
    username: str = Field(..., min_length=3, max_length=50)
    full_name: Optional[str] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class UserCreate(UserBase):
    """User creation model."""

    password: str = Field(..., min_length=8)
    plan_type: PlanType

    model_config = ConfigDict(arbitrary_types_allowed=True)


class UserResponse(BaseModel):
    """User response model."""

    id: int
    email: EmailStr
    username: str
    full_name: Optional[str] = None
    is_active: bool
    is_superuser: bool
    created_at: datetime
    updated_at: datetime
    last_login: Optional[datetime] = None

    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True)


class Token(BaseModel):
    """Token model."""

    access_token: str
    token_type: str = "bearer"
    expires_in: int
    refresh_token: Optional[str] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class TokenData(BaseModel):
    """Token data model."""

    sub: str
    exp: datetime
    refresh: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True)
